"""
Tests for win_shortcut execution module
"""

import os
import shutil
import subprocess

import pytest

from salt.exceptions import CommandExecutionError

try:
    import pythoncom
    from win32com.shell import shell

    HAS_WIN32 = True
except ImportError:
    HAS_WIN32 = False

pytestmark = [
    pytest.mark.windows_whitelisted,
    pytest.mark.skip_unless_on_windows,
    pytest.mark.skipif(not HAS_WIN32, reason="Requires Win32 libraries"),
    pytest.mark.slow_test,
]


@pytest.fixture(scope="module")
def shortcut(modules):
    return modules.shortcut


@pytest.fixture(scope="function")
def tmp_dir(tmp_path_factory):
    """
    Create a temp testing directory
    """
    test_dir = tmp_path_factory.mktemp("test_dir")
    yield test_dir
    if test_dir.exists():
        shutil.rmtree(str(test_dir))


@pytest.fixture(scope="function")
def tmp_lnk(tmp_path_factory):
    """
    Create an lnk shortcut for testing
    """
    tmp_dir = tmp_path_factory.mktemp("test_dir")
    tmp_lnk = tmp_dir / "test.lnk"
    # https://docs.microsoft.com/en-us/windows/win32/api/shobjidl_core/nn-shobjidl_core-ishelllinkw
    # http://timgolden.me.uk/python/win32_how_do_i/create-a-shortcut.html
    shortcut = pythoncom.CoCreateInstance(
        shell.CLSID_ShellLink,
        None,
        pythoncom.CLSCTX_INPROC_SERVER,
        shell.IID_IShellLink,
    )
    program = r"C:\Windows\notepad.exe"
    shortcut.SetArguments("some args")
    shortcut.SetDescription("Test description")
    shortcut.SetIconLocation(program, 0)
    shortcut.SetHotkey(1601)
    shortcut.SetPath(program)
    shortcut.SetShowCmd(1)
    shortcut.SetWorkingDirectory(os.path.dirname(program))

    persist_file = shortcut.QueryInterface(pythoncom.IID_IPersistFile)
    persist_file.Save(str(tmp_lnk), 0)

    yield tmp_lnk

    if tmp_dir.exists():
        shutil.rmtree(str(tmp_dir))


@pytest.fixture(scope="function")
def tmp_url(shortcut, tmp_path_factory):
    """
    Create a url shortcut for testing
    """
    tmp_dir = tmp_path_factory.mktemp("test_dir")
    tmp_url = tmp_dir / "test.url"
    shortcut.create(
        path=str(tmp_url),
        target="http://www.google.com",
        window_style="",
    )
    yield tmp_url

    if tmp_dir.exists():
        shutil.rmtree(str(tmp_dir))


@pytest.fixture(scope="function")
def non_lnk():
    """
    Create a file with the correct extension, but is not an actual shortcut
    """
    with pytest.helpers.temp_file("non.lnk", contents="some text") as file:
        yield file


@pytest.fixture(scope="function")
def bad_ext():
    """
    Create a temporary file with a bad file extension
    """
    with pytest.helpers.temp_file("bad.ext", contents="some text") as file:
        yield file


@pytest.fixture(scope="function")
def tmp_share():
    """
    Create a Samba Share for testing. For some reason, this is really slow...
    """
    share_dir = r"C:\Windows\Temp"
    share_name = "TmpShare"
    create_cmd = [
        "powershell",
        "-command",
        '"New-SmbShare -Name {} -Path {}" | Out-Null'.format(
            share_name, str(share_dir)
        ),
    ]
    remove_cmd = [
        "powershell",
        "-command",
        f'"Remove-SmbShare -Name {share_name} -Force" | Out-Null',
    ]
    subprocess.run(create_cmd, check=True)

    yield share_name

    subprocess.run(remove_cmd, check=True)


def test_get_missing(shortcut, tmp_dir):
    """
    Make sure that a CommandExecutionError is raised if the shortcut does NOT
    exist
    """
    fake_shortcut = tmp_dir / "fake.lnk"
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.get(path=str(fake_shortcut))
    assert "Shortcut not found" in exc.value.message


def test_get_invalid_file_extension(shortcut, bad_ext):
    """
    Make sure that a CommandExecutionError is raised if the shortcut has a non
    shortcut file extension
    """
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.get(path=str(bad_ext))
    assert exc.value.message == "Invalid file extension: .ext"


def test_get_invalid_shortcut(shortcut, non_lnk):
    """
    Make sure that a CommandExecutionError is raised if the shortcut isn't
    actually a shortcut
    """
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.get(path=str(non_lnk))
    assert "Not a valid shortcut" in exc.value.message


def test_get_lnk(shortcut, tmp_lnk):
    """
    Make sure that we return information about a valid lnk shortcut
    """
    expected = {
        "arguments": "some args",
        "description": "Test description",
        "hot_key": "Alt+Ctrl+A",
        "icon_index": 0,
        "icon_location": r"C:\Windows\notepad.exe",
        "path": str(tmp_lnk),
        "target": r"C:\Windows\notepad.exe",
        "window_style": "Normal",
        "working_dir": r"C:\Windows",
    }
    assert shortcut.get(path=str(tmp_lnk)) == expected


def test_get_url(shortcut, tmp_url):
    """
    Make sure that we return information about a valid url shortcut
    """
    expected = {
        "arguments": "",
        "description": "",
        "hot_key": "",
        "icon_index": 0,
        "icon_location": "",
        "path": str(tmp_url),
        "target": "http://www.google.com/",
        "window_style": "",
        "working_dir": "",
    }
    assert shortcut.get(path=str(tmp_url)) == expected


def test_modify_missing(shortcut, tmp_dir):
    """
    Make sure that a CommandExecutionError is raised if the shortcut does NOT
    exist
    """
    fake_shortcut = tmp_dir / "fake.lnk"
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.modify(path=str(fake_shortcut), target=r"C:\fake\path.txt")
    assert "Shortcut not found" in exc.value.message


def test_modify_invalid_file_extension(shortcut, bad_ext):
    """
    Make sure that a CommandExecutionError is raised if the shortcut has an
    invalid file extension
    """
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.modify(path=str(bad_ext), target=r"C:\fake\path.txt")
    assert exc.value.message == "Invalid file extension: .ext"


def test_modify_lnk(shortcut, tmp_lnk):
    """
    Make sure that we are able to modify an lnk shortcut
    """
    expected = {
        "arguments": "different args",
        "description": "different description",
        "hot_key": "Ctrl+Shift+B",
        "icon_index": 1,
        "icon_location": r"C:\Windows\System32\calc.exe",
        "path": str(tmp_lnk),
        "target": r"C:\Windows\System32\calc.exe",
        "window_style": "Minimized",
        "working_dir": r"C:\Windows\System32",
    }
    shortcut.modify(
        path=str(tmp_lnk),
        arguments="different args",
        description="different description",
        hot_key="Ctrl+Shift+B",
        icon_index=1,
        icon_location=r"C:\Windows\System32\calc.exe",
        target=r"C:\Windows\System32\calc.exe",
        window_style="Minimized",
        working_dir=r"C:\Windows\System32",
    )
    result = shortcut.get(path=str(tmp_lnk))
    assert result == expected


def test_modify_url(shortcut, tmp_url):
    """
    Make sure that we are able to modify a url shortcut
    """
    expected = {
        "arguments": "",
        "description": "",
        "hot_key": "",
        "icon_index": 0,
        "icon_location": "",
        "path": str(tmp_url),
        "target": "http://www.python.org/",
        "window_style": "",
        "working_dir": "",
    }
    shortcut.modify(
        path=str(tmp_url),
        target=r"www.python.org",
    )
    result = shortcut.get(path=str(tmp_url))
    assert result == expected


def test_create_invalid_file_extension(shortcut, bad_ext):
    """
    Make sure that a CommandExecutionError is raised if the shortcut file
    extension is invalid
    """
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.create(path=str(bad_ext), target=r"C:\fake\path.txt")
    assert exc.value.message == "Invalid file extension: .ext"


def test_create_existing(shortcut, tmp_lnk):
    """
    Make sure that a CommandExecutionError is raised if there is an existing
    shortcut with the same name and neither backup nor force is True
    """
    with pytest.raises(CommandExecutionError) as exc:
        shortcut.create(path=str(tmp_lnk), target=r"C:\fake\path.txt")
    assert "Found existing shortcut" in exc.value.message


def test_create_lnk(shortcut, tmp_dir):
    """
    Make sure we can create lnk type shortcut
    """
    test_link = str(os.path.join(str(tmp_dir / "test_link.lnk")))
    shortcut.create(
        path=test_link,
        arguments="create args",
        description="create description",
        hot_key="Alt+Ctrl+C",
        icon_index=0,
        icon_location=r"C:\Windows\notepad.exe",
        target=r"C:\Windows\notepad.exe",
        window_style="Normal",
        working_dir=r"C:\Windows",
    )

    expected = {
        "arguments": "create args",
        "description": "create description",
        "hot_key": "Alt+Ctrl+C",
        "icon_index": 0,
        "icon_location": r"C:\Windows\notepad.exe",
        "path": test_link,
        "target": r"C:\Windows\notepad.exe",
        "window_style": "Normal",
        "working_dir": r"C:\Windows",
    }
    result = shortcut.get(path=test_link)
    assert result == expected


@pytest.mark.slow_test
def test_create_lnk_smb_issue_61170(shortcut, tmp_dir, tmp_share):
    """
    Make sure we can create shortcuts to Samba shares
    """
    test_link = str(os.path.join(str(tmp_dir / "test_link.lnk")))
    shortcut.create(
        path=test_link,
        arguments="create args",
        description="create description",
        hot_key="Alt+Ctrl+C",
        icon_index=0,
        icon_location=r"C:\Windows\notepad.exe",
        target=rf"\\localhost\{tmp_share}",
        window_style="Normal",
        working_dir=r"C:\Windows",
    )

    expected = {
        "arguments": "create args",
        "description": "create description",
        "hot_key": "Alt+Ctrl+C",
        "icon_index": 0,
        "icon_location": r"C:\Windows\notepad.exe",
        "path": test_link,
        "target": rf"\\localhost\{tmp_share}",
        "window_style": "Normal",
        "working_dir": r"C:\Windows",
    }
    result = shortcut.get(path=test_link)
    assert result == expected


def test_create_url(shortcut, tmp_dir):
    """
    Make sure we can create url type shortcuts
    """
    test_link = str(os.path.join(str(tmp_dir / "test_link.url")))
    shortcut.create(
        path=test_link,
        target="www.google.com",
    )

    expected = {
        "arguments": "",
        "description": "",
        "hot_key": "",
        "icon_index": 0,
        "icon_location": "",
        "path": test_link,
        "target": "http://www.google.com/",
        "window_style": "",
        "working_dir": "",
    }
    result = shortcut.get(path=test_link)
    assert result == expected


def test_create_force(shortcut, tmp_lnk):
    """
    Make sure we can "force" create a shortcut if it already exists
    """
    shortcut.create(
        path=str(tmp_lnk),
        arguments="create args",
        description="create description",
        hot_key="Alt+Ctrl+C",
        icon_index=0,
        icon_location=r"C:\Windows\notepad.exe",
        target=r"C:\Windows\notepad.exe",
        window_style="Normal",
        working_dir=r"C:\Windows",
        force=True,
    )

    expected = {
        "arguments": "create args",
        "description": "create description",
        "hot_key": "Alt+Ctrl+C",
        "icon_index": 0,
        "icon_location": r"C:\Windows\notepad.exe",
        "path": str(tmp_lnk),
        "target": r"C:\Windows\notepad.exe",
        "window_style": "Normal",
        "working_dir": r"C:\Windows",
    }
    result = shortcut.get(path=str(tmp_lnk))
    assert result == expected


def test_create_backup(shortcut, tmp_lnk):
    """
    Make sure we can backup a shortcut if it already exists
    """
    shortcut.create(
        path=str(tmp_lnk),
        arguments="create args",
        description="create description",
        hot_key="Alt+Ctrl+C",
        icon_index=0,
        icon_location=r"C:\Windows\notepad.exe",
        target=r"C:\Windows\notepad.exe",
        window_style="Normal",
        working_dir=r"C:\Windows",
        backup=True,
    )

    expected = {
        "arguments": "create args",
        "description": "create description",
        "hot_key": "Alt+Ctrl+C",
        "icon_index": 0,
        "icon_location": r"C:\Windows\notepad.exe",
        "path": str(tmp_lnk),
        "target": r"C:\Windows\notepad.exe",
        "window_style": "Normal",
        "working_dir": r"C:\Windows",
    }
    result = shortcut.get(path=str(tmp_lnk))
    assert result == expected
    assert len(list(tmp_lnk.parent.glob(f"{tmp_lnk.stem}-*.lnk"))) == 1


def test_create_make_dirs(shortcut, tmp_dir):
    """
    Make sure we can create the parent directories of a shortcut if they do not
    already exist
    """
    file_shortcut = tmp_dir / "subdir" / "test.lnk"
    shortcut.create(
        path=str(file_shortcut),
        arguments="create args",
        description="create description",
        hot_key="Alt+Ctrl+C",
        icon_index=0,
        icon_location=r"C:\Windows\notepad.exe",
        target=r"C:\Windows\notepad.exe",
        window_style="Normal",
        working_dir=r"C:\Windows",
        make_dirs=True,
    )

    expected = {
        "arguments": "create args",
        "description": "create description",
        "hot_key": "Alt+Ctrl+C",
        "icon_index": 0,
        "icon_location": r"C:\Windows\notepad.exe",
        "path": str(file_shortcut),
        "target": r"C:\Windows\notepad.exe",
        "window_style": "Normal",
        "working_dir": r"C:\Windows",
    }
    result = shortcut.get(path=str(file_shortcut))
    assert result == expected



import pytest

import salt.channel.client
import salt.exceptions


def test_async_methods():
    "Validate all defined async_methods and close_methods are present"
    async_classes = [
        salt.channel.client.AsyncReqChannel,
        salt.channel.client.AsyncPubChannel,
    ]
    method_attrs = [
        "async_methods",
        "close_methods",
    ]
    for cls in async_classes:
        for attr in method_attrs:
            assert hasattr(cls, attr)
            assert isinstance(getattr(cls, attr), list)
            for name in getattr(cls, attr):
                assert hasattr(cls, name)


def test_async_pub_channel_key_overwritten_by_bad_data(minion_opts, tmp_path):
    """
    Ensure AsyncPubChannel raises a SaltClientError when it encounters a bad key.

    This bug is a bit nuanced because of how the auth module uses singletons.
    We're validating an error from salt.crypt.AsyncAuth.gen_token because of
    bad key data results in a SaltClientError. This error is handled by
    Minion._connect_minion resulting error message explaining the minion
    connection failed due to bad key data.

    https://github.com/saltstack/salt/issues/68190
    """
    minion_opts["pki_dir"] = str(tmp_path)
    minion_opts["id"] = "minion"
    minion_opts["master_ip"] = "127.0.0.1"

    # This will initialize the singleton with a valid key.
    salt.channel.client.AsyncPubChannel.factory(minion_opts, crypt="aes")

    # Now we need to overwrite the bad key with the new one. When gen_token
    # gets called a SaltClientError will ge traised
    key_path = tmp_path / "minion.pem"
    key_path.chmod(0o660)
    key_path.write_text(
        "asdfiosjaoiasdfjooaisjdfo902j0ianosdifn091091jw0edw09jcr89eq79vr"
    )
    with pytest.raises(salt.exceptions.SaltClientError):
        salt.channel.client.AsyncPubChannel.factory(minion_opts, crypt="aes")



"""
    :codeauthor: Rupesh Tare <rupesht@saltstack.com>
"""

import os

import pytest

import salt.modules.daemontools as daemontools
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {daemontools: {}}


def test_start():
    """
    Test for Starts service via daemontools
    """
    mock = MagicMock(return_value=None)
    with patch.dict(daemontools.__salt__, {"file.remove": mock}):
        mock = MagicMock(return_value="")
        with patch.object(daemontools, "_service_path", mock):
            mock = MagicMock(return_value=False)
            with patch.dict(daemontools.__salt__, {"cmd.retcode": mock}):
                assert daemontools.start("name")


def test_stop():
    """
    Test for Stops service via daemontools
    """
    mock = MagicMock(return_value=None)
    with patch.dict(daemontools.__salt__, {"file.touch": mock}):
        mock = MagicMock(return_value="")
        with patch.object(daemontools, "_service_path", mock):
            mock = MagicMock(return_value=False)
            with patch.dict(daemontools.__salt__, {"cmd.retcode": mock}):
                assert daemontools.stop("name")


def test_term():
    """
    Test for Send a TERM to service via daemontools
    """
    mock = MagicMock(return_value="")
    with patch.object(daemontools, "_service_path", mock):
        mock = MagicMock(return_value=False)
        with patch.dict(daemontools.__salt__, {"cmd.retcode": mock}):
            assert daemontools.term("name")


def test_reload_():
    """
    Test for Wrapper for term()
    """
    mock = MagicMock(return_value=None)
    with patch.object(daemontools, "term", mock):
        assert daemontools.reload_("name") is None


def test_restart():
    """
    Test for Restart service via daemontools. This will stop/start service
    """
    mock = MagicMock(return_value=False)
    with patch.object(daemontools, "stop", mock):
        assert daemontools.restart("name") == "restart False"


def test_full_restart():
    """
    Test for Calls daemontools.restart() function
    """
    mock = MagicMock(return_value=None)
    with patch.object(daemontools, "restart", mock):
        assert daemontools.restart("name") is None


def test_status():
    """
    Test for Return the status for a service via
    daemontools, return pid if running
    """
    with patch("re.search", MagicMock(return_value=1)):
        mock = MagicMock(return_value="")
        with patch.object(daemontools, "_service_path", mock):
            mock = MagicMock(return_value="name")
            with patch.dict(daemontools.__salt__, {"cmd.run_stdout": mock}):
                assert daemontools.status("name") == ""


def test_available():
    """
    Test for Returns ``True`` if the specified service
    is available, otherwise returns``False``.
    """
    mock = MagicMock(return_value=[])
    with patch.object(daemontools, "get_all", mock):
        assert not daemontools.available("name")


def test_missing():
    """
    Test for The inverse of daemontools.available.
    """
    mock = MagicMock(return_value=[])
    with patch.object(daemontools, "get_all", mock):
        assert daemontools.missing("name")


def test_get_all():
    """
    Test for Return a list of all available services
    """
    pytest.raises(CommandExecutionError, daemontools.get_all)

    with patch.object(daemontools, "SERVICE_DIR", "A"):
        mock = MagicMock(return_value="A")
        with patch.object(os, "listdir", mock):
            assert daemontools.get_all() == ["A"]



"""
Test salt.utils.zeromq
"""

import pytest
import zmq

import salt.utils.zeromq
from salt.exceptions import SaltSystemExit
from tests.support.mock import patch
from tests.support.unit import TestCase


class UtilsTestCase(TestCase):
    @pytest.mark.skipif(
        not hasattr(zmq, "IPC_PATH_MAX_LEN"),
        reason="ZMQ does not have max length support.",
    )
    def test_check_ipc_length(self):
        """
        Ensure we throw an exception if we have a too-long IPC URI
        """
        with patch("zmq.IPC_PATH_MAX_LEN", 1):
            self.assertRaises(
                SaltSystemExit, salt.utils.zeromq.check_ipc_path_max_len, "1" * 1024
            )



"""
unit tests for the stalekey engine
"""

import logging

import pytest

import salt.engines.stalekey as stalekey
from tests.support.mock import MagicMock, mock_open, patch

log = logging.getLogger(__name__)


class MockWheel:
    def __init__(self, *args, **kwargs):
        pass

    def cmd(self, *args, **kwargs):
        return True


@pytest.fixture
def configure_loader_modules():
    return {stalekey: {}}


def test__delete_keys():
    """
    Test to ensure the _delete_keys function deletes multiple keys
    """
    with patch("salt.wheel.WheelClient", MockWheel):
        stale_keys = ["minion1", "minion2"]
        minions = {
            "minion1": 1601430462.5281658,
            "minion2": 1601430462.5281658,
        }

        ret = stalekey._delete_keys(stale_keys, minions)
        assert ret == {}

        stale_keys = ["minion1"]
        minions = {
            "minion1": 1601430462.5281658,
            "minion2": 1601430462.5281658,
        }

        ret = stalekey._delete_keys(stale_keys, minions)
        assert ret == {"minion2": 1601430462.5281658}


def test__read_presence():
    """
    Test for _read_presence returning False for no error and minions presence data
    """
    presence_data = {b"minion": 1601477127.532849}
    expected = (False, {"minion": 1601477127.532849})
    with patch("os.path.exists", return_value=True):
        with patch("salt.utils.files.fopen", mock_open()):
            with patch(
                "salt.utils.msgpack.load", MagicMock(return_value=presence_data)
            ):
                ret = stalekey._read_presence("presence_file")
                assert ret == expected


def test__write_presence():
    """
    Test for _write_presence returning False, meaning no error has occured
    """
    expected = False
    minions = {
        "minion1": 1601430462.5281658,
        "minion2": 1601430462.5281658,
    }
    with patch("salt.utils.files.fopen", mock_open()):
        ret = stalekey._write_presence("presence_file", minions)
        assert ret == expected



"""
integration tests for mac_service
"""

import plistlib

import pytest

import salt.utils.files
from salt.exceptions import CommandExecutionError

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skip_if_binaries_missing("launchctl", "plutil"),
    pytest.mark.skip_unless_on_darwin,
]


@pytest.fixture(scope="module")
def service(modules):
    return modules.service


@pytest.fixture(scope="function", autouse=True)
def service_name(service):

    service_name = "com.salt.integration.test"
    service_path = "/Library/LaunchDaemons/com.salt.integration.test.plist"

    service_data = {
        "KeepAlive": True,
        "Label": service_name,
        "ProgramArguments": ["/bin/sleep", "1000"],
        "RunAtLoad": True,
    }
    with salt.utils.files.fopen(service_path, "wb") as fp:
        plistlib.dump(service_data, fp)
    service.enable(service_name)
    service.start(service_name)

    try:
        yield service_name
    finally:
        # Try to stop the service if it's running
        try:
            service.stop(service_name)
        except CommandExecutionError:
            pass
        salt.utils.files.safe_rm(service_path)


def test_show(service, service_name):
    """
    Test service.show
    """
    # Existing Service
    service_info = service.show(service_name)
    assert isinstance(service_info, dict)
    assert service_info["plist"]["Label"] == service_name

    # Missing Service
    with pytest.raises(CommandExecutionError) as exc:
        ret = service.show("spongebob")
        assert "Service not found" in str(exc.value)


def test_launchctl(service, service_name):
    """
    Test service.launchctl
    """
    # Expected Functionality
    ret = service.launchctl("error", "bootstrap", 64)
    assert ret

    ret = service.launchctl("error", "bootstrap", 64, return_stdout=True)
    assert ret == "64: unknown error code"

    # Raise an error
    with pytest.raises(CommandExecutionError) as exc:
        ret = service.launchctl("error", "bootstrap")
        assert "Failed to error service" in str(exc.value)


def test_list(service, service_name):
    """
    Test service.list
    """
    # Expected Functionality
    ret = service.list()
    assert "PID" in ret
    ret = service.list(service_name)
    assert "{" in ret

    # Service not found
    with pytest.raises(CommandExecutionError) as exc:
        ret = service.list("spongebob")
        assert "Service not found" in str(exc.value)


def test_enable(service, service_name):
    """
    Test service.enable
    """
    ret = service.enable(service_name)
    assert ret

    with pytest.raises(CommandExecutionError) as exc:
        ret = service.enable("spongebob")
        assert "Service not found" in str(exc.value)


def test_disable(service, service_name):
    """
    Test service.disable
    """
    ret = service.disable(service_name)
    assert ret

    with pytest.raises(CommandExecutionError) as exc:
        ret = service.disable("spongebob")
        assert "Service not found" in str(exc.value)


def test_start(service, service_name):
    """
    Test service.start
    Test service.stop
    Test service.status
    """
    service.stop(service_name)
    ret = service.start(service_name)
    assert ret

    with pytest.raises(CommandExecutionError) as exc:
        ret = service.start("spongebob")
        assert "Service not found" in str(exc.value)


def test_stop(service, service_name):
    """
    Test service.stop
    """
    ret = service.stop(service_name)
    assert ret

    with pytest.raises(CommandExecutionError) as exc:
        ret = service.stop("spongebob")
        assert "Service not found" in str(exc.value)

    service.start(service_name)


def test_status(service, service_name):
    """
    Test service.status
    """
    # A running service
    ret = service.status(service_name)
    assert ret

    # A stopped service
    service.stop(service_name)
    ret = service.status(service_name)
    assert not ret

    # Service not found
    ret = service.status("spongebob")
    assert not ret

    service.start(service_name)


def test_available(service, service_name):
    """
    Test service.available
    """
    ret = service.available(service_name)
    assert ret

    ret = service.available("spongebob")
    assert not ret


def test_missing(service, service_name):
    """
    Test service.missing
    """
    ret = service.missing(service_name)
    assert not ret

    ret = service.missing("spongebob")
    assert ret


def test_enabled(service, service_name):
    """
    Test service.enabled
    """
    service.disabled(service_name)
    ret = service.enabled(service_name)
    assert ret

    with pytest.raises(CommandExecutionError) as exc:
        ret = service.enabled("spongebob")
        assert "Service not found: spongebob" in str(exc.value)


def test_disabled(service, service_name):
    """
    Test service.disabled
    """
    ret = service.disabled(service_name)
    assert not ret

    ret = service.disable(service_name)
    assert ret

    ret = service.disabled(service_name)
    assert ret

    ret = service.enable(service_name)
    assert ret

    with pytest.raises(CommandExecutionError) as exc:
        ret = service.disable("spongebob")
        assert "Service not found: spongebob" in str(exc.value)


def test_get_all(service, service_name):
    """
    Test service.get_all
    """
    services = service.get_all()
    assert isinstance(services, list)
    assert service_name in services


def test_get_enabled(service, service_name):
    """
    Test service.get_enabled
    """
    services = service.get_enabled()
    assert isinstance(services, list)
    assert service_name in services


def test_service_laoded(service, service_name):
    """
    Test service.get_enabled
    """
    ret = service.loaded(service_name)
    assert ret



import logging
import shutil
import time

import pytest

import salt.config
import salt.version

try:
    import pyinotify  # pylint: disable=unused-import

    HAS_PYINOTIFY = True
except ImportError:
    HAS_PYINOTIFY = False


log = logging.getLogger(__name__)

pytestmark = [
    pytest.mark.slow_test,
    pytest.mark.skipif(HAS_PYINOTIFY is False, reason="pyinotify is not available"),
    pytest.mark.skipif(
        salt.utils.platform.is_freebsd(),
        reason="Skip on FreeBSD, IN_CREATE event is not supported",
    ),
]


@pytest.fixture
def inotify_test_path(tmp_path_factory):
    test_path = tmp_path_factory.mktemp("inotify-tests")
    try:
        yield test_path
    finally:
        shutil.rmtree(str(test_path), ignore_errors=True)


@pytest.fixture
def setup_beacons(mm_master_1_salt_cli, salt_mm_minion_1, inotify_test_path):
    start_time = time.time()
    try:
        # Add a status beacon to use for interval checks
        ret = mm_master_1_salt_cli.run(
            "beacons.add",
            "inotify",
            beacon_data=[{"files": {str(inotify_test_path): {"mask": ["create"]}}}],
            minion_tgt=salt_mm_minion_1.id,
            timeout=60,
        )
        assert ret.returncode == 0
        log.debug("Inotify beacon add returned: %s", ret.data or ret.stdout)
        assert ret.data
        assert ret.data["result"] is True
        ret = mm_master_1_salt_cli.run(
            "beacons.add",
            "status",
            beacon_data=[{"time": ["all"]}],
            minion_tgt=salt_mm_minion_1.id,
        )
        assert ret.returncode == 0
        log.debug("Status beacon add returned: %s", ret.data or ret.stdout)
        assert ret.data
        assert ret.data["result"] is True
        ret = mm_master_1_salt_cli.run(
            "beacons.list", return_yaml=False, minion_tgt=salt_mm_minion_1.id
        )
        assert ret.returncode == 0
        log.debug("Beacons list: %s", ret.data or ret.stdout)
        assert ret.data
        assert "inotify" in ret.data
        assert ret.data["inotify"] == [
            {"files": {str(inotify_test_path): {"mask": ["create"]}}}
        ]
        assert "status" in ret.data
        assert ret.data["status"] == [{"time": ["all"]}]
        yield start_time
    finally:
        # Remove the added beacons
        for beacon in ("inotify", "status"):
            mm_master_1_salt_cli.run(
                "beacons.delete", beacon, minion_tgt=salt_mm_minion_1.id
            )


@pytest.mark.slow_test
def test_beacons_duplicate_53344(
    event_listener,
    inotify_test_path,
    salt_mm_master_1,
    salt_mm_master_2,
    salt_mm_minion_1,
    setup_beacons,
):
    # We have to wait beacon first execution that would configure the inotify watch.
    # Since beacons will be executed both together, we wait for the status beacon event
    # which means that, the inotify becacon was executed too
    start_time = setup_beacons
    expected_tag = f"salt/beacon/{salt_mm_minion_1.id}/status/*"
    expected_patterns = [
        (salt_mm_master_1.id, expected_tag),
        (salt_mm_master_2.id, expected_tag),
    ]
    matched_events = event_listener.wait_for_events(
        expected_patterns,
        after_time=start_time,
        timeout=salt_mm_minion_1.config["loop_interval"] * 2 + 60,
    )
    assert matched_events.found_all_events
    log.debug("Status events received: %s", matched_events.matches)

    # Let's trigger an inotify event
    start_time = time.time()
    file_path = inotify_test_path / "tmpfile"
    file_path.write_text("")
    log.warning(
        "Test file to trigger the inotify event has been written to: %s", file_path
    )
    expected_tag = "salt/beacon/{}/inotify/{}".format(
        salt_mm_minion_1.id, inotify_test_path
    )
    expected_patterns = [
        (salt_mm_master_1.id, expected_tag),
        (salt_mm_master_2.id, expected_tag),
    ]
    matched_events = event_listener.wait_for_events(
        expected_patterns,
        after_time=start_time,
        # Now in successful case this test will get results at most in 3 loop intervals.
        # Waiting for 3 loops intervals + some seconds to the hardware stupidity.
        timeout=salt_mm_minion_1.config["loop_interval"] * 3 + 60,
    )
    assert matched_events.found_all_events
    log.debug("Inotify events received: %s", matched_events.matches)

    expected_data = {
        "path": str(file_path),
        "change": "IN_CREATE",
        "id": salt_mm_minion_1.id,
    }

    assert [(expected_tag, expected_data), (expected_tag, expected_data)] == [
        (event.tag, event.data) for event in matched_events
    ]



"""
    :codeauthor: Pedro Algarvio (pedro@algarvio.me)
    :codeauthor: Alexandru Bleotu (alexandru.bleotu@morganstanley.com)

    tests.unit.pillar_test
    ~~~~~~~~~~~~~~~~~~~~~~
"""

import logging
import os
import shutil
import tempfile
import textwrap

import pytest

import salt.config
import salt.exceptions
import salt.fileclient
import salt.utils.stringutils
from salt.utils.files import fopen
from tests.support.mock import MagicMock, patch
from tests.support.runtests import RUNTIME_VARS

log = logging.getLogger(__name__)


@pytest.fixture
def grains():
    return {}


@pytest.fixture
def tmp_pki(tmp_path):
    return str(tmp_path)


class MockFileclient:
    def __init__(self, cache_file=None, get_state=None, list_states=None):
        if cache_file is not None:
            self.cache_file = lambda *x, **y: cache_file
        if get_state is not None:
            self.get_state = lambda sls, env: get_state[sls]
        if list_states is not None:
            self.list_states = lambda *x, **y: list_states

    # pylint: disable=unused-argument,no-method-argument,method-hidden
    def cache_file(self, *args, **kwargs):
        raise NotImplementedError()

    def get_state(self, *args, **kwargs):
        raise NotImplementedError()

    def list_states(self, *args, **kwargs):
        raise NotImplementedError()

    # pylint: enable=unused-argument,no-method-argument,method-hidden


def _setup_test_topfile_sls_pillar_match(tmp_path):
    # Write a simple topfile and two pillar state files
    top_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    s = """
base:
  'phase:alpha':
    - match: pillar
    - generic
"""
    top_file.write(salt.utils.stringutils.to_bytes(s))
    top_file.flush()
    generic_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    generic_file.write(
        b"""
generic:
  key1: value1
"""
    )
    generic_file.flush()
    return {
        "top": {"path": "", "dest": top_file.name},
        "generic": {"path": "", "dest": generic_file.name},
    }


def test_pillarenv_from_saltenv():
    with patch("salt.pillar.compile_template") as compile_template:
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": {"dev": [], "base": []},
            "file_roots": {"dev": [], "base": []},
            "extension_modules": "",
            "pillarenv_from_saltenv": True,
            "fileserver_backend": "",
            "cachedir": "",
        }
        grains = {
            "os": "Ubuntu",
        }
        pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "dev")
        assert pillar.opts["saltenv"] == "dev"
        assert pillar.opts["pillarenv"] == "dev"


def _setup_test_topfile_sls(tmp_path, nodegroup_order, glob_order):
    # Write a simple topfile and two pillar state files
    top_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    s = """
base:
  group:
    - match: nodegroup
    - order: {nodegroup_order}
    - ssh
    - generic
  '*':
    - generic
  minion:
    - order: {glob_order}
    - ssh.minion
    - generic.minion
""".format(
        nodegroup_order=nodegroup_order, glob_order=glob_order
    )
    top_file.write(salt.utils.stringutils.to_bytes(s))
    top_file.flush()
    ssh_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    ssh_file.write(
        b"""
ssh:
  foo
"""
    )
    ssh_file.flush()
    ssh_minion_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    ssh_minion_file.write(
        b"""
ssh:
  bar
"""
    )
    ssh_minion_file.flush()
    generic_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    generic_file.write(
        b"""
generic:
  key1:
    - value1
    - value2
  key2:
    sub_key1: []
"""
    )
    generic_file.flush()
    generic_minion_file = tempfile.NamedTemporaryFile(dir=str(tmp_path), delete=False)
    generic_minion_file.write(
        b"""
generic:
  key1:
    - value3
  key2:
    sub_key2: []
"""
    )
    generic_minion_file.flush()

    return {
        "top": {"path": "", "dest": top_file.name},
        "ssh": {"path": "", "dest": ssh_file.name},
        "ssh.minion": {"path": "", "dest": ssh_minion_file.name},
        "generic": {"path": "", "dest": generic_file.name},
        "generic.minion": {"path": "", "dest": generic_minion_file.name},
    }


def _setup_test_include_sls(tempdir):
    top_file = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    top_file.write(
        b"""
base:
  '*':
    - order: 1
    - test.sub2
  minion:
   - order: 2
   - test
"""
    )
    top_file.flush()
    init_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    init_sls.write(
        b"""
include:
  - test.sub1
  - test.sub_wildcard*
  - .test.sub_with_init_dot
  - test/sub/with/slashes
"""
    )
    init_sls.flush()
    sub1_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    sub1_sls.write(
        b"""
foo1:
  bar1
"""
    )
    sub1_sls.flush()
    sub2_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    sub2_sls.write(
        b"""
foo2:
  bar2
"""
    )
    sub2_sls.flush()

    sub_wildcard_1_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    sub_wildcard_1_sls.write(
        b"""
foo_wildcard:
  bar_wildcard
"""
    )
    sub_wildcard_1_sls.flush()

    sub_with_init_dot_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    sub_with_init_dot_sls.write(
        b"""
sub_init_dot:
  sub_with_init_dot_worked
"""
    )
    sub_with_init_dot_sls.flush()

    sub_with_slashes_sls = tempfile.NamedTemporaryFile(dir=tempdir, delete=False)
    sub_with_slashes_sls.write(
        b"""
sub_with_slashes:
  sub_slashes_worked
"""
    )
    sub_with_slashes_sls.flush()

    return {
        "top": {"path": "", "dest": top_file.name},
        "test": {"path": "", "dest": init_sls.name},
        "test.sub1": {"path": "", "dest": sub1_sls.name},
        "test.sub2": {"path": "", "dest": sub2_sls.name},
        "test.sub_wildcard_1": {"path": "", "dest": sub_wildcard_1_sls.name},
        "test.sub_with_init_dot": {"path": "", "dest": sub_with_init_dot_sls.name},
        "test.sub.with.slashes": {"path": "", "dest": sub_with_slashes_sls.name},
    }


def test_ext_pillar_no_extra_minion_data_val_dict():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {"dev": [], "base": []},
        "file_roots": {"dev": [], "base": []},
        "extension_modules": "",
        "pillarenv_from_saltenv": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    mock_ext_pillar_func = MagicMock()
    with patch(
        "salt.loader.pillars",
        MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
    ):
        pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "dev")
    # ext pillar function doesn't have the extra_minion_data arg
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=[])),
    ):
        pillar._external_pillar_data("fake_pillar", {"arg": "foo"}, "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", arg="foo"
    )
    # ext pillar function has the extra_minion_data arg
    mock_ext_pillar_func.reset_mock()
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
    ):
        pillar._external_pillar_data("fake_pillar", {"arg": "foo"}, "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", arg="foo"
    )


def test_ext_pillar_no_extra_minion_data_val_list():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {"dev": [], "base": []},
        "file_roots": {"dev": [], "base": []},
        "extension_modules": "",
        "pillarenv_from_saltenv": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    mock_ext_pillar_func = MagicMock()
    with patch(
        "salt.loader.pillars",
        MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
    ):
        pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "dev")
    # ext pillar function doesn't have the extra_minion_data arg
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=[])),
    ):
        pillar._external_pillar_data("fake_pillar", ["foo"], "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with("mocked-minion", "fake_pillar", "foo")
    # ext pillar function has the extra_minion_data arg
    mock_ext_pillar_func.reset_mock()
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
    ):
        pillar._external_pillar_data("fake_pillar", ["foo"], "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with("mocked-minion", "fake_pillar", "foo")


def test_ext_pillar_no_extra_minion_data_val_elem():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {"dev": [], "base": []},
        "file_roots": {"dev": [], "base": []},
        "extension_modules": "",
        "pillarenv_from_saltenv": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    mock_ext_pillar_func = MagicMock()
    with patch(
        "salt.loader.pillars",
        MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
    ):
        pillar = salt.pillar.Pillar(opts, {}, "mocked-minion", "dev")
    # ext pillar function doesn't have the extra_minion_data arg
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=[])),
    ):
        pillar._external_pillar_data("fake_pillar", "fake_val", "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", "fake_val"
    )
    # ext pillar function has the extra_minion_data arg
    mock_ext_pillar_func.reset_mock()
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
    ):
        pillar._external_pillar_data("fake_pillar", "fake_val", "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", "fake_val"
    )


def test_ext_pillar_with_extra_minion_data_val_dict():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {"dev": [], "base": []},
        "file_roots": {"dev": [], "base": []},
        "extension_modules": "",
        "pillarenv_from_saltenv": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    mock_ext_pillar_func = MagicMock()
    with patch(
        "salt.loader.pillars",
        MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
    ):
        pillar = salt.pillar.Pillar(
            opts, {}, "mocked-minion", "dev", extra_minion_data={"fake_key": "foo"}
        )
    # ext pillar function doesn't have the extra_minion_data arg
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=[])),
    ):
        pillar._external_pillar_data("fake_pillar", {"arg": "foo"}, "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", arg="foo"
    )
    # ext pillar function has the extra_minion_data arg
    mock_ext_pillar_func.reset_mock()
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
    ):
        pillar._external_pillar_data("fake_pillar", {"arg": "foo"}, "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion",
        "fake_pillar",
        arg="foo",
        extra_minion_data={"fake_key": "foo"},
    )


def test_ext_pillar_with_extra_minion_data_val_list():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {"dev": [], "base": []},
        "file_roots": {"dev": [], "base": []},
        "extension_modules": "",
        "pillarenv_from_saltenv": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    mock_ext_pillar_func = MagicMock()
    with patch(
        "salt.loader.pillars",
        MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
    ):
        pillar = salt.pillar.Pillar(
            opts, {}, "mocked-minion", "dev", extra_minion_data={"fake_key": "foo"}
        )
    # ext pillar function doesn't have the extra_minion_data arg
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=[])),
    ):
        pillar._external_pillar_data("fake_pillar", ["bar"], "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with("mocked-minion", "fake_pillar", "bar")
    # ext pillar function has the extra_minion_data arg
    mock_ext_pillar_func.reset_mock()
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
    ):
        pillar._external_pillar_data("fake_pillar", ["bar"], "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", "bar", extra_minion_data={"fake_key": "foo"}
    )


def test_ext_pillar_with_extra_minion_data_val_elem():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {"dev": [], "base": []},
        "file_roots": {"dev": [], "base": []},
        "extension_modules": "",
        "pillarenv_from_saltenv": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    mock_ext_pillar_func = MagicMock()
    with patch(
        "salt.loader.pillars",
        MagicMock(return_value={"fake_ext_pillar": mock_ext_pillar_func}),
    ):
        pillar = salt.pillar.Pillar(
            opts, {}, "mocked-minion", "dev", extra_minion_data={"fake_key": "foo"}
        )
    # ext pillar function doesn't have the extra_minion_data arg
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=[])),
    ):
        pillar._external_pillar_data("fake_pillar", "bar", "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with("mocked-minion", "fake_pillar", "bar")
    # ext pillar function has the extra_minion_data arg
    mock_ext_pillar_func.reset_mock()
    with patch(
        "salt.utils.args.get_function_argspec",
        MagicMock(return_value=MagicMock(args=["extra_minion_data"])),
    ):
        pillar._external_pillar_data("fake_pillar", "bar", "fake_ext_pillar")
    mock_ext_pillar_func.assert_called_once_with(
        "mocked-minion", "fake_pillar", "bar", extra_minion_data={"fake_key": "foo"}
    )


def test_ext_pillar_first(tmp_path):
    """
    test when using ext_pillar and ext_pillar_first
    """
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "yaml",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": [],
        "extension_modules": "",
        "saltenv": "base",
        "file_roots": [],
        "ext_pillar_first": True,
        "fileserver_backend": "",
        "cachedir": "",
    }
    grains = {
        "os": "Ubuntu",
        "os_family": "Debian",
        "oscodename": "raring",
        "osfullname": "Ubuntu",
        "osrelease": "13.04",
        "kernel": "Linux",
    }

    sls_files = _setup_test_topfile_sls_pillar_match(
        str(tmp_path),
    )
    fc_mock = MockFileclient(
        cache_file=sls_files["top"]["dest"],
        list_states=["top", "ssh", "ssh.minion", "generic", "generic.minion"],
        get_state=sls_files,
    )
    with patch.object(
        salt.fileclient, "get_file_client", MagicMock(return_value=fc_mock)
    ), patch(
        "salt.pillar.Pillar.ext_pillar",
        MagicMock(
            return_value=(
                {"id": "minion", "phase": "alpha", "role": "database"},
                [],
            )
        ),
    ):
        pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
        assert pillar.compile_pillar()["generic"]["key1"] == "value1"


@patch("salt.fileclient.Client.list_states")
def test_malformed_pillar_sls(mock_list_states):
    with patch("salt.pillar.compile_template") as compile_template:
        opts = {
            "optimization_order": [0, 1, 2],
            "renderer": "json",
            "renderer_blacklist": [],
            "renderer_whitelist": [],
            "state_top": "",
            "pillar_roots": [],
            "file_roots": [],
            "extension_modules": "",
            "fileserver_backend": "",
            "cachedir": "",
        }
        grains = {
            "os": "Ubuntu",
            "os_family": "Debian",
            "oscodename": "raring",
            "osfullname": "Ubuntu",
            "osrelease": "13.04",
            "kernel": "Linux",
        }

        mock_list_states.return_value = ["foo", "blah"]
        pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
        # Mock getting the proper template files
        pillar.client.get_state = MagicMock(
            return_value={
                "dest": "/path/to/pillar/files/foo.sls",
                "source": "salt://foo.sls",
            }
        )

        # Template compilation returned a string
        compile_template.return_value = "BAHHH"
        assert pillar.render_pillar({"base": ["foo.sls"]}) == (
            {},
            ["SLS 'foo.sls' does not render to a dictionary"],
        )

        # Template compilation returned a list
        compile_template.return_value = ["BAHHH"]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == (
            {},
            ["SLS 'foo.sls' does not render to a dictionary"],
        )

        # Template compilation returned a dictionary, which is what's expected
        compile_template.return_value = {"foo": "bar"}
        assert pillar.render_pillar({"base": ["foo.sls"]}) == ({"foo": "bar"}, [])

        # Test improper includes
        compile_template.side_effect = [
            {"foo": "bar", "include": "blah"},
            {"foo2": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == (
            {"foo": "bar", "include": "blah"},
            ["Include Declaration in SLS 'foo.sls' is not formed as a list"],
        )

        # Test includes as a list, which is what's expected
        compile_template.side_effect = [
            {"foo": "bar", "include": ["blah"]},
            {"foo2": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == (
            {"foo": "bar", "foo2": "bar2"},
            [],
        )

        # Test includes as a list overriding data
        compile_template.side_effect = [
            {"foo": "bar", "include": ["blah"]},
            {"foo": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == ({"foo": "bar"}, [])

        # Test includes using empty key directive
        compile_template.side_effect = [
            {"foo": "bar", "include": [{"blah": {"key": ""}}]},
            {"foo": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == ({"foo": "bar"}, [])

        # Test includes using simple non-nested key
        compile_template.side_effect = [
            {"foo": "bar", "include": [{"blah": {"key": "nested"}}]},
            {"foo": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == (
            {"foo": "bar", "nested": {"foo": "bar2"}},
            [],
        )

        # Test includes using nested key
        compile_template.side_effect = [
            {"foo": "bar", "include": [{"blah": {"key": "nested:level"}}]},
            {"foo": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == (
            {"foo": "bar", "nested": {"level": {"foo": "bar2"}}},
            [],
        )


def test_includes_override_sls():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "json",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": {},
        "file_roots": {},
        "extension_modules": "",
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    grains = {
        "os": "Ubuntu",
        "os_family": "Debian",
        "oscodename": "raring",
        "osfullname": "Ubuntu",
        "osrelease": "13.04",
        "kernel": "Linux",
    }
    with patch("salt.pillar.compile_template") as compile_template, patch.object(
        salt.pillar.Pillar,
        "_Pillar__gather_avail",
        MagicMock(return_value={"base": ["blah", "foo"]}),
    ):

        # Test with option set to True
        opts["pillar_includes_override_sls"] = True
        pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
        # Mock getting the proper template files
        pillar.client.get_state = MagicMock(
            return_value={
                "dest": "/path/to/pillar/files/foo.sls",
                "source": "salt://foo.sls",
            }
        )

        compile_template.side_effect = [
            {"foo": "bar", "include": ["blah"]},
            {"foo": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == ({"foo": "bar2"}, [])

        # Test with option set to False
        opts["pillar_includes_override_sls"] = False
        pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
        # Mock getting the proper template files
        pillar.client.get_state = MagicMock(
            return_value={
                "dest": "/path/to/pillar/files/foo.sls",
                "source": "salt://foo.sls",
            }
        )

        compile_template.side_effect = [
            {"foo": "bar", "include": ["blah"]},
            {"foo": "bar2"},
        ]
        assert pillar.render_pillar({"base": ["foo.sls"]}) == ({"foo": "bar"}, [])


def test_topfile_order():
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "yaml",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": [],
        "extension_modules": "",
        "saltenv": "base",
        "file_roots": [],
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    grains = {
        "os": "Ubuntu",
        "os_family": "Debian",
        "oscodename": "raring",
        "osfullname": "Ubuntu",
        "osrelease": "13.04",
        "kernel": "Linux",
    }

    def _run_test(nodegroup_order, glob_order, expected):
        tempdir = tempfile.mkdtemp(dir=RUNTIME_VARS.TMP)
        try:
            sls_files = _setup_test_topfile_sls(tempdir, nodegroup_order, glob_order)
            fc_mock = MockFileclient(
                cache_file=sls_files["top"]["dest"],
                list_states=[
                    "top",
                    "ssh",
                    "ssh.minion",
                    "generic",
                    "generic.minion",
                ],
                get_state=sls_files,
            )
            with patch.object(
                salt.fileclient, "get_file_client", MagicMock(return_value=fc_mock)
            ):
                pillar = salt.pillar.Pillar(opts, grains, "mocked-minion", "base")
                # Make sure that confirm_top.confirm_top returns True
                pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True
                assert pillar.compile_pillar()["ssh"] == expected
        finally:
            shutil.rmtree(tempdir, ignore_errors=True)

    # test case where glob match happens second and therefore takes
    # precedence over nodegroup match.
    _run_test(nodegroup_order=1, glob_order=2, expected="bar")
    # test case where nodegroup match happens second and therefore takes
    # precedence over glob match.
    _run_test(nodegroup_order=2, glob_order=1, expected="foo")


def test_relative_include(tmp_path):
    join = os.path.join
    with fopen(join(str(tmp_path), "top.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                base:
                  '*':
                    - includer
                    - simple_includer
                    - includes.with.more.depth
            """
            ),
            file=f,
        )
    includer_dir = str(tmp_path / "includer")
    os.makedirs(includer_dir)
    with fopen(join(includer_dir, "init.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                include:
                  - .this
                  - includer.that
            """
            ),
            file=f,
        )
    with fopen(join(includer_dir, "this.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                this:
                    is all good
            """
            ),
            file=f,
        )
    with fopen(join(includer_dir, "that.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                that:
                    is also all good
            """
            ),
            file=f,
        )

    with fopen(str(tmp_path / "simple_includer.sls"), "w") as simpleincluder:
        print(
            textwrap.dedent(
                """
                include:
                  - .simple
                  - super_simple
            """
            ),
            file=simpleincluder,
        )
    with fopen(str(tmp_path / "simple.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                simple:
                  simon
            """
            ),
            file=f,
        )
    with fopen(str(tmp_path / "super_simple.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                super simple:
                  a caveman
            """
            ),
            file=f,
        )

    depth_dir = tmp_path / "includes" / "with" / "more"
    os.makedirs(str(depth_dir))
    with fopen(str(depth_dir / "depth.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                include:
                  - .ramble
                  - includes.with.more.doors

                mordor:
                    has dark depths
            """
            ),
            file=f,
        )

    with fopen(str(depth_dir / "ramble.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                found:
                    my precious
            """
            ),
            file=f,
        )

    with fopen(str(depth_dir / "doors.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                mojo:
                    bad risin'
            """
            ),
            file=f,
        )
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "yaml",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "top.sls",
        "pillar_roots": {"base": [str(tmp_path)]},
        "extension_modules": "",
        "saltenv": "base",
        "file_roots": [],
        "file_ignore_regex": None,
        "file_ignore_glob": None,
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    grains = {
        "os": "Ubuntu",
        "os_family": "Debian",
        "oscodename": "raring",
        "osfullname": "Ubuntu",
        "osrelease": "13.04",
        "kernel": "Linux",
    }
    pillar = salt.pillar.Pillar(opts, grains, "minion", "base")
    # Make sure that confirm_top.confirm_top returns True
    pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True

    # Act
    compiled_pillar = pillar.compile_pillar()

    # Assert
    assert compiled_pillar["this"] == "is all good"
    assert compiled_pillar["that"] == "is also all good"
    assert compiled_pillar["simple"] == "simon"
    assert compiled_pillar["super simple"] == "a caveman"
    assert compiled_pillar["mordor"] == "has dark depths"
    assert compiled_pillar["found"] == "my precious"
    assert compiled_pillar["mojo"] == "bad risin'"


def test_missing_include(tmp_path):
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "yaml",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "top.sls",
        "pillar_roots": {"base": [str(tmp_path)]},
        "extension_modules": "",
        "saltenv": "base",
        "file_roots": [],
        "file_ignore_regex": None,
        "file_ignore_glob": None,
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    grains = {
        "os": "Ubuntu",
        "os_family": "Debian",
        "oscodename": "raring",
        "osfullname": "Ubuntu",
        "osrelease": "13.04",
        "kernel": "Linux",
    }

    with fopen(str(tmp_path / "top.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                base:
                  '*':
                    - simple_include
            """
            ),
            file=f,
        )
    include_dir = tmp_path / "simple_include"
    os.makedirs(include_dir)
    with fopen(str(include_dir / "init.sls"), "w") as f:
        print(
            textwrap.dedent(
                """
                include:
                  - simple_include.missing_include
                simple_include: is ok
                """
            ),
            file=f,
        )

    pillar = salt.pillar.Pillar(opts, grains, "minion", "base")
    # Make sure that confirm_top.confirm_top returns True
    pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True

    # Act
    compiled_pillar = pillar.compile_pillar()

    # Assert
    assert compiled_pillar["simple_include"] == "is ok"
    assert "_errors" in compiled_pillar
    assert "simple_include.missing_include" in compiled_pillar["_errors"][0]


def test_get_opts_in_pillar_override_call(minion_opts, grains):
    mock_get_extra_minion_data = MagicMock(return_value={})
    with patch(
        "salt.pillar.RemotePillarMixin.get_ext_pillar_extra_minion_data",
        mock_get_extra_minion_data,
    ):

        salt.pillar.RemotePillar(minion_opts, grains, "mocked-minion", "dev")
    call_opts = dict(minion_opts, saltenv="dev")
    mock_get_extra_minion_data.assert_called_once_with(call_opts)


def test_multiple_keys_in_opts_added_to_pillar(grains, tmp_pki):
    opts = {
        "pki_dir": tmp_pki,
        "id": "minion",
        "master_uri": "tcp://127.0.0.1:4505",
        "__role": "minion",
        "keysize": 2048,
        "renderer": "json",
        "path_to_add": "fake_data",
        "path_to_add2": {"fake_data2": ["fake_data3", "fake_data4"]},
        "pass_to_ext_pillars": ["path_to_add", "path_to_add2"],
    }
    pillar = salt.pillar.RemotePillar(opts, grains, "mocked-minion", "dev")
    assert pillar.extra_minion_data == {
        "path_to_add": "fake_data",
        "path_to_add2": {"fake_data2": ["fake_data3", "fake_data4"]},
    }


def test_subkey_in_opts_added_to_pillar(minion_opts, grains):
    opts = dict(
        minion_opts,
        path_to_add2={
            "fake_data5": "fake_data6",
            "fake_data2": ["fake_data3", "fake_data4"],
        },
        pass_to_ext_pillars=["path_to_add2:fake_data5"],
    )
    pillar = salt.pillar.RemotePillar(opts, grains, "mocked-minion", "dev")
    assert pillar.extra_minion_data == {"path_to_add2": {"fake_data5": "fake_data6"}}


def test_non_existent_leaf_opt_in_add_to_pillar(minion_opts, grains):
    pillar = salt.pillar.RemotePillar(minion_opts, grains, "mocked-minion", "dev")
    assert pillar.pillar_override == {}


def test_non_existent_intermediate_opt_in_add_to_pillar(minion_opts, grains):
    pillar = salt.pillar.RemotePillar(minion_opts, grains, "mocked-minion", "dev")
    assert pillar.pillar_override == {}


def test_malformed_add_to_pillar(minion_opts, grains):
    opts = dict(minion_opts, pass_to_ext_pillars=MagicMock())
    with pytest.raises(salt.exceptions.SaltClientError) as excinfo:
        salt.pillar.RemotePillar(opts, grains, "mocked-minion", "dev")
        assert (
            excinfo.exception.strerror == "'pass_to_ext_pillars' config is malformed."
        )


def test_pillar_file_client_master_remote(tmp_pki, grains):
    """
    Test condition where local file_client and use_master_when_local option
    returns a remote file client.
    """
    mocked_minion = MagicMock()
    opts = {
        "pki_dir": tmp_pki,
        "id": "minion",
        "master_uri": "tcp://127.0.0.1:4505",
        "__role": "minion",
        "keysize": 2048,
        "file_client": "local",
        "use_master_when_local": True,
        "pillar_cache": None,
    }
    pillar = salt.pillar.get_pillar(opts, grains, mocked_minion)
    assert type(pillar) == salt.pillar.RemotePillar
    assert type(pillar) != salt.pillar.PillarCache


def test_pillar_send_extra_minion_data_from_config(tmp_pki, grains):
    opts = {
        "pki_dir": tmp_pki,
        "id": "minion",
        "master_uri": "tcp://127.0.0.1:4505",
        "__role": "minion",
        "keysize": 2048,
        "renderer": "json",
        "pillarenv": "fake_pillar_env",
        "path_to_add": "fake_data",
        "path_to_add2": {
            "fake_data5": "fake_data6",
            "fake_data2": ["fake_data3", "fake_data4"],
        },
        "pass_to_ext_pillars": ["path_to_add"],
    }
    mock_channel = MagicMock(
        crypted_transfer_decode_dictentry=MagicMock(return_value={})
    )
    with patch(
        "salt.channel.client.AsyncReqChannel.factory",
        MagicMock(return_value=mock_channel),
    ):
        pillar = salt.pillar.RemotePillar(opts, grains, "mocked_minion", "fake_env")

    ret = pillar.compile_pillar()
    mock_channel.crypted_transfer_decode_dictentry.assert_called_once_with(
        {
            "cmd": "_pillar",
            "ver": "2",
            "id": "mocked_minion",
            "grains": {},
            "saltenv": "fake_env",
            "pillarenv": "fake_pillar_env",
            "pillar_override": {},
            "extra_minion_data": {"path_to_add": "fake_data"},
        },
        dictkey="pillar",
    )


def test_include(tmp_path):
    opts = {
        "optimization_order": [0, 1, 2],
        "renderer": "yaml",
        "renderer_blacklist": [],
        "renderer_whitelist": [],
        "state_top": "",
        "pillar_roots": [],
        "extension_modules": "",
        "saltenv": "base",
        "file_roots": [],
        "fileserver_backend": "roots",
        "cachedir": "",
    }
    grains = {
        "os": "Ubuntu",
        "os_family": "Debian",
        "oscodename": "raring",
        "osfullname": "Ubuntu",
        "osrelease": "13.04",
        "kernel": "Linux",
    }
    sls_files = _setup_test_include_sls(str(tmp_path))
    fc_mock = MockFileclient(
        cache_file=sls_files["top"]["dest"],
        get_state=sls_files,
        list_states=[
            "top",
            "test.init",
            "test.sub1",
            "test.sub2",
            "test.sub_wildcard_1",
            "test.sub_with_init_dot",
            "test.sub.with.slashes",
        ],
    )
    with patch.object(
        salt.fileclient, "get_file_client", MagicMock(return_value=fc_mock)
    ):
        pillar = salt.pillar.Pillar(opts, grains, "minion", "base")
        # Make sure that confirm_top.confirm_top returns True
        pillar.matchers["confirm_top.confirm_top"] = lambda *x, **y: True
        compiled_pillar = pillar.compile_pillar()
        assert compiled_pillar["foo_wildcard"] == "bar_wildcard"
        assert compiled_pillar["foo1"] == "bar1"
        assert compiled_pillar["foo2"] == "bar2"
        assert compiled_pillar["sub_with_slashes"] == "sub_slashes_worked"
        assert compiled_pillar["sub_init_dot"] == "sub_with_init_dot_worked"


def test_compile_pillar_memory_cache(master_opts):
    master_opts.update({"pillar_cache_backend": "memory", "pillar_cache_ttl": 3600})

    pillar = salt.pillar.PillarCache(
        master_opts,
        grains,
        "mocked_minion",
        "fake_env",
        pillarenv="base",
    )

    with patch(
        "salt.pillar.PillarCache.fetch_pillar",
        side_effect=[{"foo": "bar"}, {"foo": "baz"}],
    ):
        # Run once for pillarenv base
        ret = pillar.compile_pillar()
        expected_cache = {"base": {"foo": "bar"}}
        assert "mocked_minion" in pillar.cache
        assert pillar.cache["mocked_minion"] == expected_cache

        # Run a second time for pillarenv base
        ret = pillar.compile_pillar()
        expected_cache = {"base": {"foo": "bar"}}
        assert "mocked_minion" in pillar.cache
        assert pillar.cache["mocked_minion"] == expected_cache

        # Change the pillarenv
        pillar.pillarenv = "dev"

        # Run once for pillarenv dev
        ret = pillar.compile_pillar()
        expected_cache = {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
        assert "mocked_minion" in pillar.cache
        assert pillar.cache["mocked_minion"] == expected_cache

        # Run a second time for pillarenv dev
        ret = pillar.compile_pillar()
        expected_cache = {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
        assert "mocked_minion" in pillar.cache
        assert pillar.cache["mocked_minion"] == expected_cache


def test_compile_pillar_disk_cache(master_opts, grains):
    master_opts.update({"pillar_cache_backend": "disk", "pillar_cache_ttl": 3600})

    pillar = salt.pillar.PillarCache(
        master_opts,
        grains,
        "mocked_minion",
        "fake_env",
        pillarenv="base",
    )

    with patch("salt.utils.cache.CacheDisk._write", MagicMock()):
        with patch(
            "salt.pillar.PillarCache.fetch_pillar",
            side_effect=[{"foo": "bar"}, {"foo": "baz"}],
        ):
            # Run once for pillarenv base
            ret = pillar.compile_pillar()
            expected_cache = {"mocked_minion": {"base": {"foo": "bar"}}}
            assert pillar.cache._dict == expected_cache

            # Run a second time for pillarenv base
            ret = pillar.compile_pillar()
            expected_cache = {"mocked_minion": {"base": {"foo": "bar"}}}
            assert pillar.cache._dict == expected_cache

            # Change the pillarenv
            pillar.pillarenv = "dev"

            # Run once for pillarenv dev
            ret = pillar.compile_pillar()
            expected_cache = {
                "mocked_minion": {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
            }
            assert pillar.cache._dict == expected_cache

            # Run a second time for pillarenv dev
            ret = pillar.compile_pillar()
            expected_cache = {
                "mocked_minion": {"base": {"foo": "bar"}, "dev": {"foo": "baz"}}
            }
            assert pillar.cache._dict == expected_cache


def test_remote_pillar_bad_return(grains, tmp_pki):
    opts = {
        "pki_dir": tmp_pki,
        "id": "minion",
        "master_uri": "tcp://127.0.0.1:4505",
        "__role": "minion",
        "keysize": 2048,
        "saltenv": "base",
        "pillarenv": "base",
    }
    pillar = salt.pillar.RemotePillar(opts, grains, "mocked-minion", "dev")

    async def crypted_transfer_mock():
        return ""

    pillar.channel.crypted_transfer_decode_dictentry = crypted_transfer_mock
    with pytest.raises(salt.exceptions.SaltClientError):
        pillar.compile_pillar()


async def test_async_remote_pillar_bad_return(grains, tmp_pki):
    opts = {
        "pki_dir": tmp_pki,
        "id": "minion",
        "master_uri": "tcp://127.0.0.1:4505",
        "__role": "minion",
        "keysize": 2048,
        "saltenv": "base",
        "pillarenv": "base",
    }
    pillar = salt.pillar.AsyncRemotePillar(opts, grains, "mocked-minion", "dev")

    async def crypted_transfer_mock():
        return ""

    pillar.channel.crypted_transfer_decode_dictentry = crypted_transfer_mock
    with pytest.raises(salt.exceptions.SaltClientError):
        await pillar.compile_pillar()



#!/usr/bin/env python3
# see issue: https://gitlab.com/saltstack/open/salt-pkg/-/issues/19
import contextlib
import multiprocessing
import os
import pathlib
import sys

import _pyio
import tiamatpip.cli
import tiamatpip.configure
import tiamatpip.utils

import salt.scripts
import salt.utils.platform

AVAIL = (
    "minion",
    "master",
    "call",
    "api",
    "cloud",
    "cp",
    "extend",
    "key",
    "proxy",
    "pip",
    "run",
    "shell",
    "spm",
    "ssh",
    "support",
    "syndic",
    "python",
)


if "TIAMAT_PIP_PYPATH" in os.environ:
    PIP_PATH = pathlib.Path(os.environ["TIAMAT_PIP_PYPATH"]).resolve()
elif not sys.platform.startswith("win"):
    PIP_PATH = pathlib.Path(f"{os.sep}opt", "saltstack", "salt", "pypath")
else:
    PIP_PATH = pathlib.Path(os.getenv("LocalAppData"), "salt", "pypath")
with contextlib.suppress(PermissionError):
    PIP_PATH.mkdir(mode=0o755, parents=True, exist_ok=True)
tiamatpip.configure.set_user_base_path(PIP_PATH)


def py_shell():
    if not sys.platform.startswith("win"):
        # optional, will allow Up/Down/History in the console
        import readline
    import code

    variables = globals().copy()
    variables.update(locals())
    shell = code.InteractiveConsole(variables)
    shell.interact()


def python_runtime():
    import traceback

    # extract the absolute script path to alter sys.path and specific dunder variables
    script = pathlib.Path(sys.argv[2]).expanduser().resolve()
    sys.path.insert(0, str(script.parent))

    # update passed args so they don't start with "<binary> python"
    sys.argv[:] = sys.argv[2:]
    exec_locals = {"__name__": "__main__", "__file__": str(script), "__doc__": None}
    with open(script, encoding="utf-8") as rfh:
        try:
            exec(rfh.read(), exec_locals)
        except Exception:
            traceback.print_exc()
            sys.exit(1)


def redirect(argv):
    """
    Change the args and redirect to another salt script
    """
    if len(argv) < 2:
        msg = "Must pass in a salt command, available commands are:"
        for cmd in AVAIL:
            msg += f"\n{cmd}"
        print(msg, file=sys.stderr, flush=True)
        sys.exit(1)
    cmd = sys.argv[1]
    if cmd == "shell":
        py_shell()
        return
    if cmd == "python":
        if len(argv) < 3:
            msg = "Must pass script location to this command"
            print(msg, file=sys.stderr, flush=True)
            sys.exit(1)

        python_runtime()
        return
    if tiamatpip.cli.should_redirect_argv(argv):
        tiamatpip.cli.process_pip_argv(argv)
        return
    if cmd not in AVAIL:
        # Fall back to the salt command
        args = ["salt"]
        s_fun = salt.scripts.salt_main
    else:
        args = [f"salt-{cmd}"]
        sys.argv.pop(1)
        s_fun = getattr(salt.scripts, f"salt_{cmd}")
    args.extend(argv[1:])
    with tiamatpip.utils.patched_sys_argv(args):
        s_fun()


if __name__ == "__main__":
    multiprocessing.freeze_support()
    redirect(sys.argv)



"""
Digicert Pillar Certificates

This module will only return pillar data if the ``digicert`` runner module has
already been used to create certificates.

To configure this module, set ``digicert`` to ``True`` in the ``ext_pillar``
section of your ``master`` configuration file:

.. code-block:: yaml

    ext_pillar:
      - digicert: True
"""

import logging

import salt.cache
import salt.syspaths as syspaths

__virtualname__ = "digicert"
log = logging.getLogger(__name__)


def __virtual__():
    """
    No special requirements outside of Salt itself
    """
    return __virtualname__


def ext_pillar(minion_id, pillar, conf):
    """
    Return an existing set of certificates
    """
    cache = salt.cache.Cache(__opts__, syspaths.CACHE_DIR)

    ret = {}
    dns_names = cache.fetch("digicert/minions", minion_id)

    for dns_name in dns_names:
        data = cache.fetch("digicert/domains", dns_name)
        ret[dns_name] = data
        del ret[dns_name]["csr"]
    return {"digicert": ret}



"""
    :codeauthor: `Anthony Shaw <anthonyshaw@apache.org>`

    tests.pytests.unit.cloud.clouds.test_gce
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

import collections

import pytest

from salt.cloud.clouds import gce
from salt.exceptions import SaltCloudSystemExit
from salt.utils.versions import Version
from tests.support.mock import MagicMock
from tests.support.mock import __version__ as mock_version
from tests.support.mock import call, patch


@pytest.fixture
def configure_loader_modules():

    return {
        gce: {
            "show_instance": MagicMock(),
            "__active_provider_name__": "",
            "__utils__": {
                "cloud.fire_event": MagicMock(),
                "cloud.filter_event": MagicMock(),
            },
            "__opts__": {
                "sock_dir": True,
                "transport": True,
                "providers": {
                    "my-google-cloud": {
                        "gce": {
                            "project": "daenerys-cloud",
                            "service_account_email_address": (
                                "dany@targaryen.westeros.cloud"
                            ),
                            "service_account_private_key": "/home/dany/PRIVKEY.pem",
                            "driver": "gce",
                            "ssh_interface": "public_ips",
                        }
                    }
                },
            },
        }
    }


@pytest.fixture(
    params=[
        {"expected": "", "image": ""},
        {"expected": None, "image": None},
        {"expected": "debian-10", "image": "debian-10"},
    ]
)
def config_image(request):
    return request.param["expected"], request.param["image"]


@pytest.fixture(
    params=[
        {"expected": None, "label": "{}"},
        {"expected": {"mylabel": "myvalue"}, "label": "{'mylabel': 'myvalue'}"},
    ]
)
def config_labels(request):
    return request.param["expected"], request.param["label"]


@pytest.fixture(
    params=[
        {
            "expected": collections.namedtuple("Location", "name")("chicago"),
            "location": collections.namedtuple("Location", "name")("chicago"),
        },
    ]
)
def config_location(request):
    return request.param["expected"], request.param["location"]


@pytest.fixture(
    params=[
        {
            "expected": {"items": [{"key": "salt-cloud-profile", "value": None}]},
            "metadata": {},
        },
        {
            "expected": {
                "items": [
                    {"key": "mykey", "value": "myvalue"},
                    {"key": "salt-cloud-profile", "value": None},
                ]
            },
            "metadata": "{'mykey': 'myvalue'}",
        },
    ]
)
def config_metadata(request):
    return request.param["expected"], request.param["metadata"]


@pytest.fixture(
    params=[
        {"expected": "mynetwork", "network": "mynetwork"},
    ]
)
def config_network(request):
    return request.param["expected"], request.param["network"]


@pytest.fixture(
    params=[
        {"expected": "e2-standard-2", "size": "e2-standard-2"},
    ]
)
def config_size(request):
    return request.param["expected"], request.param["size"]


@pytest.fixture(
    params=[
        {"expected": "mysubnetwork", "subnetwork": "mysubnetwork"},
    ]
)
def config_subnetwork(request):
    return request.param["expected"], request.param["subnetwork"]


@pytest.fixture(
    params=[
        {"expected": None, "tag": "{}"},
        {"expected": ["mytag", "myvalue"], "tag": "['mytag', 'myvalue']"},
    ]
)
def config_tags(request):
    return request.param["expected"], request.param["tag"]


@pytest.fixture
def config(
    config_image,
    config_labels,
    config_location,
    config_metadata,
    config_network,
    config_size,
    config_subnetwork,
    config_tags,
):
    expected_image, image = config_image
    expected_labels, labels = config_labels
    expected_location, location = config_location
    expected_metadata, metadata = config_metadata
    expected_network, network = config_network
    expected_size, size = config_size
    expected_subnetwork, subnetwork = config_subnetwork
    expected_tags, tags = config_tags
    expected_call_kwargs = {
        "ex_disk_type": "pd-standard",
        "ex_metadata": expected_metadata,
        "ex_accelerator_count": 42,
        "name": "new",
        "ex_service_accounts": None,
        "external_ip": "ephemeral",
        "ex_accelerator_type": "foo",
        "ex_tags": expected_tags,
        "ex_labels": expected_labels,
        "ex_disk_auto_delete": True,
        "ex_network": expected_network,
        "ex_disks_gce_struct": None,
        "ex_preemptible": False,
        "ex_can_ip_forward": False,
        "ex_on_host_maintenance": "TERMINATE",
        "location": expected_location,
        "ex_subnetwork": expected_subnetwork,
        "image": expected_image,
        "size": expected_size,
    }
    config = {
        "name": "new",
        "driver": "gce",
        "profile": None,
        "size": size,
        "image": image,
        "location": location,
        "ex_accelerator_type": "foo",
        "ex_accelerator_count": 42,
        "network": network,
        "subnetwork": subnetwork,
        "ex_labels": labels,
        "tags": tags,
        "metadata": metadata,
    }
    return expected_call_kwargs, config


@pytest.fixture
def fake_libcloud_2_3_0():
    with patch("salt.cloud.clouds.gce.LIBCLOUD_VERSION_INFO", (2, 3, 0)):
        yield


@pytest.fixture
def fake_libcloud_2_5_0():
    with patch("salt.cloud.clouds.gce.LIBCLOUD_VERSION_INFO", (2, 5, 0)):
        yield


@pytest.fixture
def conn():
    def return_first(*args, **kwargs):
        return args[0]

    with patch("salt.cloud.clouds.gce.get_conn", autospec=True) as fake_conn:
        fake_addy = MagicMock()
        fake_addy.extra = {}
        fake_addy.region.name = "fnord town"
        fake_conn.return_value.ex_create_address.return_value = fake_addy
        fake_conn.return_value.ex_get_network.side_effect = return_first
        fake_conn.return_value.ex_get_image.side_effect = return_first
        fake_conn.return_value.ex_get_zone.side_effect = return_first
        fake_conn.return_value.ex_get_size.side_effect = return_first
        yield fake_conn.return_value


@pytest.fixture
def fake_conf_provider():
    with patch("salt.config.is_provider_configured", autospec=True) as fake_conf:
        yield fake_conf


def test_destroy_call():
    """
    Tests that a SaltCloudSystemExit is raised when trying to call destroy
    with --function or -f.
    """
    pytest.raises(
        SaltCloudSystemExit, gce.destroy, vm_name="kings_landing", call="function"
    )


@pytest.mark.skipif(
    gce.HAS_LIBCLOUD is False, reason="apache-libcloud is not installed"
)
def test_fail_virtual_libcloud_version_too_old():
    # Missing deps
    with patch("libcloud.__version__", "2.4.0"):
        v = gce.__virtual__()
        assert v == (False, "The salt-cloud GCE driver requires apache-libcloud>=2.5.0")


@pytest.mark.skipif(
    gce.HAS_LIBCLOUD is False, reason="apache-libcloud is not installed"
)
def test_fail_virtual_missing_deps():
    # Missing deps
    with patch("salt.config.check_driver_dependencies", return_value=False):
        v = gce.__virtual__()
        assert v is False


@pytest.mark.skipif(
    gce.HAS_LIBCLOUD is False, reason="apache-libcloud is not installed"
)
def test_fail_virtual_deps_missing_config():
    with patch("salt.config.check_driver_dependencies", return_value=True), patch(
        "salt.config.is_provider_configured", return_value=False
    ):
        v = gce.__virtual__()
        assert v is False


def test_import():
    """
    Test that the module picks up installed deps
    """
    with patch("salt.config.check_driver_dependencies", return_value=True) as p:
        get_deps = gce.get_dependencies()
        assert get_deps is True
        if Version(mock_version) >= Version("2.0.0"):
            p.assert_called_once()


@pytest.mark.parametrize(
    "active_provider", [("fnord", "fnord"), (None, "gce"), ("", "gce")]
)
def test_get_configured_provider_should_pass_expected_args(
    active_provider, fake_conf_provider
):
    """
    gce delegates the behavior to config.is_provider_configured, and should
    pass on expected args.
    """
    provider_name, expected_provider = active_provider
    with patch(
        "salt.cloud.clouds.gce._get_active_provider_name",
        autospec=True,
        return_value=provider_name,
    ):
        gce.get_configured_provider()
    fake_conf_provider.assert_called_with(
        gce.__opts__,
        expected_provider,
        ("project", "service_account_email_address", "service_account_private_key"),
    )


def test_get_configured_provider_should_return_expected_result(fake_conf_provider):
    """
    Currently get_configured_provider should simply return whatever
    comes back from config.is_provider_configured, no questions asked.
    """
    expected_result = object()
    fake_conf_provider.return_value = expected_result

    actual_result = gce.get_configured_provider()

    assert actual_result is expected_result


def test_request_instance_with_accelerator(config, conn):
    """
    Test requesting an instance with GCE accelerators
    """
    expected_call_kwargs, vm_config = config
    gce.request_instance(vm_config)
    conn.create_node.assert_called_once_with(**expected_call_kwargs)


def test_create_address_should_fire_creating_and_created_events_with_expected_args(
    conn,
):
    region = MagicMock()
    region.name = "antarctica"
    kwargs = {"name": "bob", "region": region, "address": "123 Easy Street"}
    expected_args = {
        "name": "bob",
        "region": {"name": "antarctica"},
        "address": "123 Easy Street",
    }
    expected_creating_call = call(
        "event",
        "create address",
        "salt/cloud/address/creating",
        args=expected_args,
        sock_dir=gce.__opts__["sock_dir"],
        transport=gce.__opts__["transport"],
    )
    expected_created_call = call(
        "event",
        "created address",
        "salt/cloud/address/created",
        args=expected_args,
        sock_dir=gce.__opts__["sock_dir"],
        transport=gce.__opts__["transport"],
    )

    gce.create_address(kwargs, "function")

    gce.__utils__["cloud.fire_event"].assert_has_calls(
        [expected_creating_call, expected_created_call]
    )


def test_create_address_passes_correct_args_to_ex_create_address(
    conn, fake_libcloud_2_3_0
):
    """
    Test create_address
    """
    expected_name = "name mcnameface"
    expected_region = MagicMock(name="fnord")
    expected_address = "addresss mcaddressface"
    expected_call_args = (expected_name, expected_region, expected_address)

    kwargs = {
        "name": expected_name,
        "region": expected_region,
        "address": expected_address,
    }
    gce.create_address(kwargs, "function")

    conn.ex_create_address.assert_called_once_with(*expected_call_args)



# Python libs

import logging

import pytest

# Salt libs
import salt.beacons.network_settings as network_settings
from tests.support.mock import MagicMock, patch

try:
    # ipdb_interfaces_view requires pyroute2 >= 0.7.1
    from pyroute2 import NDB
    from pyroute2.ndb.compat import ipdb_interfaces_view

    HAS_NDB = True
except ImportError:
    HAS_NDB = False

try:
    # IPDB support may be dropped in future pyroute2 releases
    from pyroute2 import IPDB

    HAS_IPDB = True
except ImportError:
    HAS_IPDB = False


log = logging.getLogger(__name__)


class MockIPClass:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def by_name(self):
        return {}


@pytest.fixture
def configure_loader_modules():
    return {network_settings: {"__context__": {}, "__salt__": {}}}


def test_non_list_config():
    config = {}

    ret = network_settings.validate(config)

    assert ret == (False, "Configuration for network_settings beacon must be a list.")


def test_empty_config():
    config = [{}]

    ret = network_settings.validate(config)

    assert ret == (True, "Valid beacon configuration")


def test_interface():
    config = [{"interfaces": {"enp14s0u1u2": {"promiscuity": None}}}]
    LAST_STATS = network_settings._copy_interfaces_info(
        {"enp14s0u1u2": {"family": "0", "promiscuity": "0", "group": "0"}}
    )

    NEW_STATS = network_settings._copy_interfaces_info(
        {"enp14s0u1u2": {"family": "0", "promiscuity": "1", "group": "0"}}
    )

    ret = network_settings.validate(config)
    assert ret == (True, "Valid beacon configuration")

    with patch.object(network_settings, "LAST_STATS", {}), patch.object(
        network_settings, "IP", MockIPClass
    ), patch(
        "salt.beacons.network_settings._copy_interfaces_info",
        MagicMock(side_effect=[LAST_STATS, NEW_STATS]),
    ):
        ret = network_settings.beacon(config)
        assert ret == []

        ret = network_settings.beacon(config)
        _expected = [
            {
                "interface": "enp14s0u1u2",
                "tag": "enp14s0u1u2",
                "change": {"promiscuity": "1"},
            }
        ]
        assert ret == _expected


def test_interface_no_change():
    config = [{"interfaces": {"enp14s0u1u2": {"promiscuity": None}}}]
    LAST_STATS = network_settings._copy_interfaces_info(
        {"enp14s0u1u2": {"family": "0", "promiscuity": "0", "group": "0"}}
    )

    NEW_STATS = network_settings._copy_interfaces_info(
        {"enp14s0u1u2": {"family": "0", "promiscuity": "0", "group": "0"}}
    )

    ret = network_settings.validate(config)
    assert ret == (True, "Valid beacon configuration")

    with patch.object(network_settings, "LAST_STATS", {}), patch.object(
        network_settings, "IP", MockIPClass
    ), patch(
        "salt.beacons.network_settings._copy_interfaces_info",
        MagicMock(side_effect=[LAST_STATS, NEW_STATS]),
    ):
        ret = network_settings.beacon(config)
        assert ret == []

        ret = network_settings.beacon(config)
        assert ret == []


def test_wildcard_interface():
    config = [{"interfaces": {"en*": {"promiscuity": None}}}]
    LAST_STATS = network_settings._copy_interfaces_info(
        {"enp14s0u1u2": {"family": "0", "promiscuity": "0", "group": "0"}}
    )

    NEW_STATS = network_settings._copy_interfaces_info(
        {"enp14s0u1u2": {"family": "0", "promiscuity": "1", "group": "0"}}
    )

    ret = network_settings.validate(config)
    assert ret == (True, "Valid beacon configuration")

    with patch.object(network_settings, "LAST_STATS", {}), patch.object(
        network_settings, "IP", MockIPClass
    ), patch(
        "salt.beacons.network_settings._copy_interfaces_info",
        MagicMock(side_effect=[LAST_STATS, NEW_STATS]),
    ):
        ret = network_settings.beacon(config)
        assert ret == []

        ret = network_settings.beacon(config)
        _expected = [
            {
                "interface": "enp14s0u1u2",
                "tag": "enp14s0u1u2",
                "change": {"promiscuity": "1"},
            }
        ]
        assert ret == _expected


@pytest.mark.skipif(HAS_IPDB is False, reason="pyroute2.IPDB not available, skipping")
def test_interface_dict_fields_old():
    with IPDB() as ipdb:
        for attr in network_settings.ATTRS:
            # ipdb.interfaces is a dict-like object, that
            # contains interface definitions. Interfaces can
            # be referenced both with indices and names.
            #
            # ipdb.interfaces[1] is an interface with index 1,
            # that is the loopback interface.
            assert attr in ipdb.interfaces[1]


@pytest.mark.skipif(
    HAS_NDB is False, reason="pyroute2.ndb.compat not yet available, skipping"
)
def test_interface_dict_fields_new():
    with NDB() as ndb:
        # ndb provides dict-like objects for all the RTNL entities
        # upon requests, like:
        #
        #   ndb.interfaces["lo"]
        #   ndb.interfaces[{"target": "netns01", "ifname": "lo"}]
        #   ndb.addresses["127.0.0.1/8"]
        #
        # but for our case is important that NDB provides listings of
        # RTNL entities as sets of named tuples:
        #
        #   for record in ndb.interfaces:
        #       print(record.index, record.ifname)
        #
        # pyroute2.ndb.compat.ipdb_interfaces_view() translates
        # this view into a dict tree that resembles IPDB layout.
        #
        # beacon might use this translation to read the NDB info and at
        # the same time not to saturate memory with multiple objects.
        view = ipdb_interfaces_view(ndb)
        for attr in network_settings.ATTRS:
            assert attr in view["lo"]



"""
Regression test for unsafe YAML loading in junos module.

This test ensures that the junos.get_table() function uses a safe YAML loader
and does not allow arbitrary code execution.

CVE/Security Issue: The junos module was using yamlordereddictloader.Loader
which extends yaml.Loader (unsafe). It should use yamlordereddictloader.SafeLoader.
"""

import pytest
import yaml

try:
    import yamlordereddictloader

    HAS_YAMLORDEREDDICTLOADER = True
except ImportError:
    HAS_YAMLORDEREDDICTLOADER = False


pytestmark = [
    pytest.mark.skipif(
        not HAS_YAMLORDEREDDICTLOADER,
        reason="The yamlordereddictloader module is required",
    ),
]


def test_junos_module_uses_safe_yaml_loader():
    """
    Regression test to ensure junos module uses SafeLoader, not Loader.

    This test will:
    - FAIL if salt/modules/junos.py uses yamlordereddictloader.Loader (UNSAFE)
    - PASS if salt/modules/junos.py uses yamlordereddictloader.SafeLoader (SAFE)
    """
    import inspect

    import salt.modules.junos as junos

    # Get the source code of the get_table function
    source = inspect.getsource(junos.get_table)

    # Check that SafeLoader is used, not Loader
    if "yamlordereddictloader.Loader" in source and "SafeLoader" not in source.replace(
        "yamlordereddictloader.Loader", ""
    ):
        pytest.fail(
            "SECURITY VULNERABILITY: junos.get_table() uses yamlordereddictloader.Loader!\n\n"
            "The unsafe Loader can execute arbitrary Python code from YAML files.\n\n"
            "FIX: Change yamlordereddictloader.Loader to yamlordereddictloader.SafeLoader\n"
            "in salt/modules/junos.py (around line 1861)"
        )

    # Verify SafeLoader is actually used
    assert "yamlordereddictloader.SafeLoader" in source, (
        "Expected to find yamlordereddictloader.SafeLoader in junos.get_table(). "
        "Make sure the safe loader is being used."
    )


def test_yamlordereddictloader_safeloader_blocks_code_execution():
    """
    Test that yamlordereddictloader.SafeLoader blocks arbitrary code execution.

    This demonstrates the correct, secure behavior that should be used.
    """
    # Malicious YAML attempting to execute os.system
    malicious_yaml = "!!python/object/apply:os.system ['echo this should not execute']"

    # SafeLoader should reject this and raise ConstructorError
    with pytest.raises(yaml.constructor.ConstructorError):
        yaml.load(malicious_yaml, Loader=yamlordereddictloader.SafeLoader)


def test_yamlordereddictloader_loader_inheritance():
    """
    Verify the inheritance chain showing why Loader is unsafe.
    """
    # yamlordereddictloader.Loader extends yaml.Loader (UNSAFE)
    assert issubclass(
        yamlordereddictloader.Loader, yaml.Loader
    ), "yamlordereddictloader.Loader should extend yaml.Loader (unsafe)"

    # yamlordereddictloader.SafeLoader extends yaml.SafeLoader (SAFE)
    assert issubclass(
        yamlordereddictloader.SafeLoader, yaml.SafeLoader
    ), "yamlordereddictloader.SafeLoader should extend yaml.SafeLoader (safe)"


def test_ordered_dict_functionality_with_safeloader():
    """
    Verify that SafeLoader still provides OrderedDict functionality.

    This ensures that switching to SafeLoader doesn't break the intended
    functionality of maintaining key order.
    """
    test_yaml = """
TableTest:
    key1: value1
    key2: value2
    key3: value3
"""

    # Load with SafeLoader
    result = yaml.load(test_yaml, Loader=yamlordereddictloader.SafeLoader)

    # Verify it loaded successfully
    assert "TableTest" in result
    assert result["TableTest"]["key1"] == "value1"

    # Verify it returns OrderedDict (the whole point of yamlordereddictloader)
    from collections import OrderedDict

    assert isinstance(result, OrderedDict) or isinstance(result, dict)
    assert isinstance(result["TableTest"], OrderedDict) or isinstance(
        result["TableTest"], dict
    )


def test_compare_unsafe_vs_safe_loader():
    """
    Direct comparison showing the difference between unsafe and safe loaders.
    """
    # Safe YAML that should work with both loaders
    safe_yaml = """
config:
    host: example.com
    port: 8080
    enabled: true
"""

    # Load with both loaders - should work fine
    result_unsafe = yaml.load(safe_yaml, Loader=yamlordereddictloader.Loader)
    result_safe = yaml.load(safe_yaml, Loader=yamlordereddictloader.SafeLoader)

    # Both should produce the same result for safe YAML
    assert result_unsafe == result_safe

    # Malicious YAML with Python object constructor
    malicious_yaml = "!!python/object/apply:os.system ['ls']"

    # Safe Loader will raise an exception (correct behavior)
    with pytest.raises(yaml.constructor.ConstructorError):
        yaml.load(malicious_yaml, Loader=yamlordereddictloader.SafeLoader)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import pytest

import salt.states.ssh_auth as ssh_auth
from tests.support.mock import MagicMock, call, patch


@pytest.fixture
def configure_loader_modules():
    return {ssh_auth: {"__env__": "base"}}


@pytest.fixture()
def enc_types():
    # This list is from sshd manual page of openssh 8.7
    return [
        "sk-ecdsa-sha2-nistp256@openssh.com",
        "ecdsa-sha2-nistp256",
        "ecdsa-sha2-nistp384",
        "ecdsa-sha2-nistp521",
        "sk-ssh-ed25519@openssh.com",
        "ssh-ed25519",
        "ssh-dss",
        "ssh-rsa",
    ]


def test__present_test(enc_types):
    """
    Test to verify if many encryption key types are accepted by ssh_auth._present_test
    """
    user = "user"
    key = "1ABCDXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
    comment = "user@example.internal"
    name_list = []
    expected_calls = []
    for enc in enc_types:
        name_list.append(f"{enc} {key} {comment}")
        expected_calls.append(
            call(
                user,
                key,
                enc,
                comment,
                [],
                config=".ssh/authorized_keys",
                fingerprint_hash_type=None,
            )
        )

    mock_check_key = MagicMock(return_value="update")

    with patch.dict(ssh_auth.__salt__, {"ssh.check_key": mock_check_key}):
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            for name in name_list:
                ret = None, f"Key {key} for user {user} is set to be updated"
                assert (
                    ssh_auth._present_test(
                        user, name, enc, comment, [], "", ".ssh/authorized_keys", None
                    )
                    == ret
                )
            mock_check_key.assert_has_calls(expected_calls)


def test__absent_test(enc_types):
    """
    Test to verify if many encryption key types are accepted by ssh_auth._absent_test
    """
    user = "user"
    key = "1ABCDXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
    comment = "user@example.internal"
    name_list = []
    expected_calls = []
    for enc in enc_types:
        name_list.append(f"{enc} {key} {comment}")
        expected_calls.append(
            call(
                user,
                key,
                enc,
                comment,
                [],
                config=".ssh/authorized_keys",
                fingerprint_hash_type=None,
            )
        )

    mock_check_key = MagicMock(return_value="update")

    with patch.dict(ssh_auth.__salt__, {"ssh.check_key": mock_check_key}):
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            for name in name_list:
                ret = None, f"Key {key} for user {user} is set for removal"
                assert (
                    ssh_auth._absent_test(
                        user, name, enc, comment, [], "", ".ssh/authorized_keys", None
                    )
                    == ret
                )
            mock_check_key.assert_has_calls(expected_calls)


def test_present():
    """
    Test to verifies that the specified SSH key
    is present for the specified user.
    """
    name = "sshkeys"
    user = "root"
    source = "salt://ssh_keys/id_rsa.pub"

    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    mock = MagicMock(return_value="exists")
    mock_data = MagicMock(side_effect=["replace", "new"])
    with patch.dict(
        ssh_auth.__salt__, {"ssh.check_key": mock, "ssh.set_auth_key": mock_data}
    ):
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            comt = "The authorized host key sshkeys is already present for user root"
            ret.update({"comment": comt})
            assert ssh_auth.present(name, user, source) == ret

        with patch.dict(ssh_auth.__opts__, {"test": False}):
            comt = "The authorized host key sshkeys for user root was updated"
            ret.update({"comment": comt, "changes": {name: "Updated"}})
            assert ssh_auth.present(name, user, source) == ret

            comt = "The authorized host key sshkeys for user root was added"
            ret.update({"comment": comt, "changes": {name: "New"}})
            assert ssh_auth.present(name, user, source) == ret


def test_present_with_source():
    """
    Test to check if ssh.set_auth_key_from_file is called with the correct parameters
    """
    user = "user"
    source = "salt://fakefile"
    config = ".ssh/authorized_keys"

    mock_cp_get_url = MagicMock(return_value=True)
    mock_auth_key_from_file = MagicMock(return_value="new")

    with patch.dict(
        ssh_auth.__salt__,
        {
            "cp.get_url": mock_cp_get_url,
            "ssh.set_auth_key_from_file": mock_auth_key_from_file,
        },
    ):
        with patch.dict(ssh_auth.__opts__, {"test": False}):
            name = "Call Without Options"
            ret = {
                "name": name,
                "changes": {name: "New"},
                "result": True,
                "comment": f"The authorized host key {name} for user {user} was added",
            }
            assert ssh_auth.present(name, user, source=source) == ret
            mock_auth_key_from_file.assert_called_with(
                "user",
                "salt://fakefile",
                config=".ssh/authorized_keys",
                saltenv="base",
                fingerprint_hash_type=None,
                options=None,
            )
            name = "Call With Options"
            ret = {
                "name": name,
                "changes": {name: "New"},
                "result": True,
                "comment": f"The authorized host key {name} for user {user} was added",
            }
            assert (
                ssh_auth.present(name, user, source=source, options=["no-pty"]) == ret
            )
            mock_auth_key_from_file.assert_called_with(
                "user",
                "salt://fakefile",
                config=".ssh/authorized_keys",
                saltenv="base",
                fingerprint_hash_type=None,
                options=["no-pty"],
            )


def test_present_different_key_types_are_accepted(enc_types):
    """
    Test to verify if many encryption key types are accepted by ssh_auth.present
    """
    user = "user"
    key = "1ABCDXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
    comment = "user@example.internal"
    name_list = []
    expected_calls = []
    for enc in enc_types:
        name_list.append(f"{enc} {key} {comment}")
        expected_calls.append(
            call(
                user,
                key,
                enc,
                comment,
                [],
                "",
                ".ssh/authorized_keys",
                None,
            )
        )

    with patch.object(
        ssh_auth, "_present_test", return_value=[None, "Key is to be added"]
    ) as mock__present_test:
        # it's tested with `test: True` because this already tests the fullkey regexp
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            for name in name_list:
                ret = {
                    "name": name,
                    "changes": {},
                    "result": None,
                    "comment": "Key is to be added",
                }
                assert ssh_auth.present(name, user) == ret
            mock__present_test.assert_has_calls(expected_calls)


def test_absent():
    """
    Test to verifies that the specified SSH key is absent.
    """
    name = "sshkeys"
    user = "root"
    source = "salt://ssh_keys/id_rsa.pub"

    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    mock = MagicMock(
        side_effect=["User authorized keys file not present", "Key removed"]
    )
    mock_up = MagicMock(side_effect=["update", "updated"])
    with patch.dict(
        ssh_auth.__salt__, {"ssh.rm_auth_key": mock, "ssh.check_key": mock_up}
    ):
        with patch.dict(ssh_auth.__opts__, {"test": True}):
            comt = "Key sshkeys for user root is set for removal"
            ret.update({"comment": comt})
            assert ssh_auth.absent(name, user, source) == ret

            comt = "Key is already absent"
            ret.update({"comment": comt, "result": True})
            assert ssh_auth.absent(name, user, source) == ret

        with patch.dict(ssh_auth.__opts__, {"test": False}):
            comt = "User authorized keys file not present"
            ret.update({"comment": comt, "result": False})
            assert ssh_auth.absent(name, user, source) == ret

            comt = "Key removed"
            ret.update({"comment": comt, "result": True, "changes": {name: "Removed"}})
            assert ssh_auth.absent(name, user, source) == ret


def test_absent_with_source():
    """
    Test to check if ssh.rm_auth_key_from_file is called with the correct parameters
    """
    user = "user"
    source = "salt://fakefile"
    config = ".ssh/authorized_keys"

    mock_cp_get_url = MagicMock(return_value=True)
    mock_auth_key_from_file = MagicMock(return_value="Key removed")

    with patch.dict(
        ssh_auth.__salt__,
        {
            "cp.get_url": mock_cp_get_url,
            "ssh.rm_auth_key_from_file": mock_auth_key_from_file,
        },
    ):
        with patch.dict(ssh_auth.__opts__, {"test": False}):
            name = "Call Absent With Source"
            ret = {
                "name": name,
                "changes": {name: "Removed"},
                "result": True,
                "comment": "Key removed",
            }
            assert ssh_auth.absent(name, user, source=source) == ret
            mock_auth_key_from_file.assert_called_with(
                "user",
                "salt://fakefile",
                ".ssh/authorized_keys",
                saltenv="base",
                fingerprint_hash_type=None,
            )


def test_absent_with_different_key_types_are_accepted(enc_types):
    """
    Test to verify if many encription key types are accepted
    """
    user = "user"
    key = "1ABCDXXXXXXXXXXXXXXXXXXXXXXXXXXXX="
    comment = "user@example.internal"
    name_list = []
    expected_calls = []
    for enc in enc_types:
        name_list.append(f"{enc} {key} {comment}")
        expected_calls.append(
            call(
                user,
                key,
                config=".ssh/authorized_keys",
                fingerprint_hash_type=None,
            )
        )

    mock_rm_test = MagicMock(return_value="Key removed")

    with patch.dict(ssh_auth.__salt__, {"ssh.rm_auth_key": mock_rm_test}):
        # it's tested with `test: False` because the fullkey regexp is only in the actual execution
        with patch.dict(ssh_auth.__opts__, {"test": False}):
            for name in name_list:
                ret = {
                    "name": name,
                    "changes": {key: "Removed"},
                    "result": True,
                    "comment": "Key removed",
                }
                assert ssh_auth.absent(name, user) == ret
            mock_rm_test.assert_has_calls(expected_calls)


def test_manage():
    """
    Test to verifies that the specified SSH key is absent.
    """
    user = "root"
    ret = {"name": "", "changes": {}, "result": None, "comment": ""}

    mock_rm = MagicMock(
        side_effect=["User authorized keys file not present", "Key removed"]
    )
    mock_up = MagicMock(side_effect=["update", "updated"])
    mock_set = MagicMock(side_effect=["replace", "new"])
    mock_keys = MagicMock(
        return_value={
            "somekey": {
                "enc": "ssh-rsa",
                "comment": "user@host",
                "options": [],
                "fingerprint": "b7",
            }
        }
    )
    with patch.dict(
        ssh_auth.__salt__,
        {
            "ssh.rm_auth_key": mock_rm,
            "ssh.set_auth_key": mock_set,
            "ssh.check_key": mock_up,
            "ssh.auth_keys": mock_keys,
        },
    ):
        with patch("salt.states.ssh_auth.present") as call_mocked_present:
            mock_present = {"comment": "", "changes": {}, "result": None}
            call_mocked_present.return_value = mock_present
            with patch.dict(ssh_auth.__opts__, {"test": True}):
                assert ssh_auth.manage("sshid", ["somekey"], user) == ret

                comt = "somekey Key set for removal"
                ret.update({"comment": comt})
                assert ssh_auth.manage("sshid", [], user) == ret

        with patch("salt.states.ssh_auth.present") as call_mocked_present:
            mock_present = {"comment": "", "changes": {}, "result": True}
            call_mocked_present.return_value = mock_present
            with patch.dict(ssh_auth.__opts__, {"test": False}):
                ret = {"name": "", "changes": {}, "result": True, "comment": ""}
                assert ssh_auth.manage("sshid", ["somekey"], user) == ret

                with patch("salt.states.ssh_auth.absent") as call_mocked_absent:
                    mock_absent = {"comment": "Key removed"}
                    call_mocked_absent.return_value = mock_absent
                    ret.update(
                        {
                            "comment": "",
                            "result": True,
                            "changes": {"somekey": "Key removed"},
                        }
                    )
                    assert ssh_auth.manage("sshid", ["addkey"], user) == ret

        # add a key
        with patch("salt.states.ssh_auth.present") as call_mocked_present:
            mock_present = {
                "comment": (
                    f"The authorized host key newkey for user {user} was added"
                ),
                "changes": {"newkey": "New"},
                "result": True,
            }
            call_mocked_present.return_value = mock_present
            with patch.dict(ssh_auth.__opts__, {"test": False}):
                ret = {
                    "name": "",
                    "changes": {"newkey": "New"},
                    "result": True,
                    "comment": "",
                }
                assert ssh_auth.manage("sshid", ["newkey", "somekey"], user) == ret



import datetime
import os
import time

import salt.fileserver
import salt.utils.files


def test_diff_with_diffent_keys():
    """
    Test that different maps are indeed reported different
    """
    map1 = {"file1": 1234}
    map2 = {"file2": 1234}
    assert salt.fileserver.diff_mtime_map(map1, map2) is True


def test_diff_with_diffent_values():
    """
    Test that different maps are indeed reported different
    """
    map1 = {"file1": 12345}
    map2 = {"file1": 1234}
    assert salt.fileserver.diff_mtime_map(map1, map2) is True


def test_whitelist():
    opts = {
        "fileserver_backend": ["roots", "git", "s3fs", "hgfs", "svn"],
        "extension_modules": "",
    }
    fs = salt.fileserver.Fileserver(opts)
    assert sorted(fs.servers.whitelist) == sorted(
        ["git", "gitfs", "hg", "hgfs", "svn", "svnfs", "roots", "s3fs"]
    ), fs.servers.whitelist


def test_future_file_list_cache_file_ignored(tmp_path):
    opts = {
        "fileserver_backend": ["roots"],
        "cachedir": tmp_path,
        "extension_modules": "",
    }

    back_cachedir = os.path.join(tmp_path, "file_lists/roots")
    os.makedirs(os.path.join(back_cachedir))

    # Touch a couple files
    for filename in ("base.p", "foo.txt"):
        with salt.utils.files.fopen(os.path.join(back_cachedir, filename), "wb") as _f:
            if filename == "base.p":
                _f.write(b"\x80")

    # Set modification time to file list cache file to 1 year in the future
    now = datetime.datetime.utcnow()
    future = now + datetime.timedelta(days=365)
    mod_time = time.mktime(future.timetuple())
    os.utime(os.path.join(back_cachedir, "base.p"), (mod_time, mod_time))

    list_cache = os.path.join(back_cachedir, "base.p")
    w_lock = os.path.join(back_cachedir, ".base.w")
    ret = salt.fileserver.check_file_list_cache(opts, "files", list_cache, w_lock)
    assert (
        ret[1] is True
    ), "Cache file list cache file is not refreshed when future modification time"


def test_file_server_url_escape(tmp_path):
    (tmp_path / "srv").mkdir()
    (tmp_path / "srv" / "salt").mkdir()
    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "bar").write_text("Bad file")
    fileroot = str(tmp_path / "srv" / "salt")
    badfile = str(tmp_path / "foo" / "bar")
    opts = {
        "fileserver_backend": ["roots"],
        "extension_modules": "",
        "optimization_order": [0, 1],
        "file_roots": {
            "base": [fileroot],
        },
        "file_ignore_regex": "",
        "file_ignore_glob": "",
    }
    fs = salt.fileserver.Fileserver(opts)
    ret = fs.find_file(
        "salt://|..\\..\\..\\foo/bar",
        "base",
    )
    assert ret == {"path": "", "rel": ""}


def test_file_server_serve_url_escape(tmp_path):
    (tmp_path / "srv").mkdir()
    (tmp_path / "srv" / "salt").mkdir()
    (tmp_path / "foo").mkdir()
    (tmp_path / "foo" / "bar").write_text("Bad file")
    fileroot = str(tmp_path / "srv" / "salt")
    badfile = str(tmp_path / "foo" / "bar")
    opts = {
        "fileserver_backend": ["roots"],
        "extension_modules": "",
        "optimization_order": [0, 1],
        "file_roots": {
            "base": [fileroot],
        },
        "file_ignore_regex": "",
        "file_ignore_glob": "",
        "file_buffer_size": 2048,
    }
    fs = salt.fileserver.Fileserver(opts)
    ret = fs.serve_file(
        {
            "path": "salt://|..\\..\\..\\foo/bar",
            "saltenv": "base",
            "loc": 0,
        }
    )
    assert ret == {"data": "", "dest": ""}



"""
Unit tests for the postgres_local_cache.
"""

import json

import pytest

import salt.returners.postgres_local_cache as postgres_local_cache
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules(tmp_path):
    return {
        postgres_local_cache: {
            "__opts__": {
                "cachedir": str(tmp_path / "cache_dir"),
                "keep_jobs_seconds": 3600,
            }
        }
    }


def test_returner():
    """
    Tests that the returner function
    """
    load = {
        "tgt_type": "glob",
        "fun_args": [],
        "jid": "20200108221839189167",
        "return": True,
        "retcode": 0,
        "success": True,
        "tgt": "minion",
        "cmd": "_return",
        "_stamp": "2020-01-08T22:18:39.197774",
        "arg": [],
        "fun": "test.ping",
        "id": "minion",
    }

    expected = {"return": "True", "retcode": 0, "success": True}

    connect_mock = MagicMock()
    with patch.object(postgres_local_cache, "_get_conn", connect_mock):
        postgres_local_cache.returner(load)

        return_val = None
        for call in connect_mock.mock_calls:
            for arg in call.args:
                if isinstance(arg, tuple):
                    for val in arg:
                        if isinstance(val, str) and "return" in val:
                            return_val = json.loads(val)

        assert return_val is not None, None
        assert return_val == expected


def test_returner_unicode_exception():
    """
    Tests that the returner function
    """
    return_val = "Trüe"

    load = {
        "tgt_type": "glob",
        "fun_args": [],
        "jid": "20200108221839189167",
        "return": return_val,
        "retcode": 0,
        "success": True,
        "tgt": "minion",
        "cmd": "_return",
        "_stamp": "2020-01-08T22:18:39.197774",
        "arg": [],
        "fun": "test.ping",
        "id": "minion",
    }

    expected = {"return": "Trüe", "retcode": 0, "success": True}

    connect_mock = MagicMock()
    with patch.object(postgres_local_cache, "_get_conn", connect_mock):
        postgres_local_cache.returner(load)

        return_val = None
        search_string = "return"
        for call in connect_mock.mock_calls:
            for arg in call.args:
                if isinstance(arg, tuple):
                    for val in arg:
                        if isinstance(val, str):
                            if search_string in val:
                                return_val = json.loads(val)

        assert return_val is not None, None
        assert return_val == expected



"""
    :codeauthor: Jayesh Kariya <jayeshk@saltstack.com>
"""

import os

import pytest

import salt.states.ports as ports
from salt.exceptions import SaltInvocationError
from tests.support.mock import MagicMock, patch


class MockModule:
    __module__ = "A"


class MockContext:
    __context__ = {"ports.install_error": "salt"}


class MockSys:
    def __init__(self):
        self.modules = {"A": MockContext()}


@pytest.fixture
def configure_loader_modules():
    return {ports: {"sys": MockSys()}}


def test_installed():
    """
    Test to verify that the desired port is installed,
    and that it was compiled with the desired options.
    """
    name = "security/nmap"
    options = [{"IPV6": "on"}]

    ret = {"name": name, "result": None, "comment": "", "changes": {}}

    mock = MagicMock(side_effect=SaltInvocationError)
    with patch.dict(ports.__salt__, {"ports.showconfig": mock}):
        comt = (
            "Unable to get configuration for {}. Port name may "
            "be invalid, or ports tree may need to be updated. "
            "Error message: ".format(name)
        )
        ret.update({"comment": comt, "result": False})
        assert ports.installed(name) == ret

    mock = MagicMock(return_value={})
    mock_lst = MagicMock(return_value={"origin": {"origin": name}})
    with patch.dict(
        ports.__salt__, {"ports.showconfig": mock, "pkg.list_pkgs": mock_lst}
    ):
        comt = "security/nmap is already installed"
        ret.update({"comment": comt, "result": True})
        assert ports.installed(name) == ret

        comt = (
            "security/nmap does not have any build options, yet options were specified"
        )
        ret.update({"comment": comt, "result": False})
        assert ports.installed(name, options) == ret

        mock_dict = MagicMock(return_value={"origin": {"origin": "salt"}})
        with patch.dict(ports.__salt__, {"pkg.list_pkgs": mock_dict}):
            with patch.dict(ports.__opts__, {"test": True}):
                comt = f"{name} will be installed"
                ret.update({"comment": comt, "result": None})
                assert ports.installed(name) == ret

    mock = MagicMock(return_value={"salt": {"salt": "salt"}})
    mock_dict = MagicMock(return_value={"origin": {"origin": "salt"}})
    mock_f = MagicMock(return_value=False)
    mock_t = MagicMock(return_value=True)
    with patch.dict(
        ports.__salt__,
        {
            "ports.showconfig": mock,
            "pkg.list_pkgs": mock_dict,
            "ports.config": mock_f,
            "ports.rmconfig": mock_t,
        },
    ):
        with patch.dict(ports.__opts__, {"test": True}):
            comt = "The following options are not available for security/nmap: IPV6"
            ret.update({"comment": comt, "result": False})
            assert ports.installed(name, options) == ret

            comt = "security/nmap will be installed with the default build options"
            ret.update({"comment": comt, "result": None})
            assert ports.installed(name) == ret

        with patch.dict(ports.__opts__, {"test": False}):
            comt = "Unable to set options for security/nmap"
            ret.update({"comment": comt, "result": False})
            assert ports.installed(name, [{"salt": "salt"}]) == ret

            with patch.object(os.path, "isfile", mock_t):
                with patch.object(os.path, "isdir", mock_t):
                    comt = "Unable to clear options for security/nmap"
                    ret.update({"comment": comt, "result": False})
                    assert ports.installed(name) == ret

            with patch.dict(
                ports.__salt__,
                {
                    "ports.config": mock_t,
                    "ports.install": mock_t,
                    "test.ping": MockModule(),
                },
            ):
                comt = "Failed to install security/nmap. Error message:\nsalt"
                ret.update({"comment": comt, "result": False, "changes": True})
                assert ports.installed(name, [{"salt": "salt"}]) == ret



"""
    salt.serializers.plist
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~
    .. versionadded:: 3001

    Implements plist serializer.

    Wrapper around plistlib.
"""

import logging
import plistlib

from salt.serializers import DeserializationError, SerializationError

log = logging.getLogger(__name__)

__all__ = ["deserialize", "serialize", "available"]

available = True


def deserialize(stream_or_string, **options):
    """
    Deserialize any string or stream like object into a Python data structure.

    :param stream_or_string: stream or string to deserialize.
    :param options: options given to lower plist module.

    :returns: Deserialized data structure.
    """
    try:
        if not isinstance(stream_or_string, (bytes, str)):
            log.trace("Using plistlib.load to deserialize.")
            return plistlib.load(stream_or_string, **options)

        if isinstance(stream_or_string, str):
            log.trace("Need to encode plist string.")
            stream_or_string = stream_or_string.encode("utf-8")

        log.trace("Using plistlib.loads to deserialize.")
        return plistlib.loads(stream_or_string, **options)
    except Exception as error:  # pylint: disable=broad-except
        raise DeserializationError(error)


def serialize(value, **options):
    """
    Serialize Python data to plist. To create a binary plist pass
    ``fmt: FMT_BINARY`` as an option.

    :param obj: the data structure to serialize
    :param options: options given to lower plist module.

    :returns: bytes of serialized plist.
    """
    fmt = options.pop("fmt", None)
    # add support for serializing to binary.
    if fmt == "FMT_BINARY":
        log.trace("Adding plistlib.FMT_BINARY to options.")
        options["fmt"] = plistlib.FMT_BINARY

    try:
        if "fp" in options:
            log.trace("Using plistlib.dump to serialize.")
            return plistlib.dump(value, **options)

        log.trace("Using plistlib.dumps to serialize.")
        return plistlib.dumps(value, **options)
    except Exception as error:  # pylint: disable=broad-except
        raise SerializationError(error)



"""
Network NTP
===========

.. versionadded:: 2016.11.0

Manage the configuration of NTP peers and servers on the network devices through the NAPALM proxy.

:codeauthor: Mircea Ulinic <ping@mirceaulinic.net> & Jerome Fleury <jf@cloudflare.com>
:maturity:   new
:depends:    napalm
:platform:   unix

Dependencies
------------
- Requires netaddr_ to be installed: `pip install netaddr` to check if IP
  Addresses are correctly specified
- Requires dnspython_ to be installed: `pip install dnspython` to resolve the
  nameserver entities (in case the user does not configure the peers/servers
  using their IP addresses)
- :mod:`NAPALM proxy minion <salt.proxy.napalm>`
- :mod:`NTP operational and configuration management module <salt.modules.napalm_ntp>`

.. _netaddr: https://pythonhosted.org/netaddr/
.. _dnspython: http://www.dnspython.org/
"""

import logging

import salt.utils.napalm

try:
    from netaddr import IPAddress
    from netaddr.core import AddrFormatError

    HAS_NETADDR = True
except ImportError:
    HAS_NETADDR = False

try:
    import dns.resolver  # pylint: disable=no-name-in-module

    HAS_DNSRESOLVER = True
except ImportError:
    HAS_DNSRESOLVER = False

# ----------------------------------------------------------------------------------------------------------------------
# state properties
# ----------------------------------------------------------------------------------------------------------------------

__virtualname__ = "netntp"

log = logging.getLogger(__name__)

# ----------------------------------------------------------------------------------------------------------------------
# global variables
# ----------------------------------------------------------------------------------------------------------------------

# ----------------------------------------------------------------------------------------------------------------------
# property functions
# ----------------------------------------------------------------------------------------------------------------------


def __virtual__():
    """
    NAPALM library must be installed for this module to work and run in a (proxy) minion.
    """
    return salt.utils.napalm.virtual(__opts__, __virtualname__, __file__)


# ----------------------------------------------------------------------------------------------------------------------
# helper functions -- will not be exported
# ----------------------------------------------------------------------------------------------------------------------


def _default_ret(name):

    ret = {"name": name, "changes": {}, "result": False, "comment": ""}
    return ret


def _retrieve_ntp_peers():
    """Retrieves configured NTP peers"""

    return __salt__["ntp.peers"]()


def _retrieve_ntp_servers():
    """Retrieves configured NTP servers"""

    return __salt__["ntp.servers"]()


def _check(peers):
    """Checks whether the input is a valid list of peers and transforms domain names into IP Addresses"""

    if not isinstance(peers, list):
        return False

    for peer in peers:
        if not isinstance(peer, str):
            return False

    if (
        not HAS_NETADDR
    ):  # if does not have this lib installed, will simply try to load what user specified
        # if the addresses are not correctly specified, will trow error when loading the actual config
        return True

    ip_only_peers = []
    for peer in peers:
        try:
            ip_only_peers.append(str(IPAddress(peer)))  # append the str value
        except AddrFormatError:
            # if not a valid IP Address
            # will try to see if it is a nameserver and resolve it
            if not HAS_DNSRESOLVER:
                continue  # without the dns resolver cannot populate the list of NTP entities based on their nameserver
                # so we'll move on
            dns_reply = []
            try:
                # try to see if it is a valid NS
                dns_reply = dns.resolver.query(peer)
            except dns.resolver.NoAnswer:
                # no a valid DNS entry either
                return False
            for dns_ip in dns_reply:
                ip_only_peers.append(str(dns_ip))

    peers = ip_only_peers

    return True


def _clean(lst):

    return [elem for elem in lst if elem]


def _set_ntp_peers(peers):
    """Calls ntp.set_peers."""

    return __salt__["ntp.set_peers"](*peers, commit=False)


def _set_ntp_servers(servers):
    """Calls ntp.set_servers."""

    return __salt__["ntp.set_servers"](*servers, commit=False)


def _delete_ntp_peers(peers):
    """Calls ntp.delete_peers."""

    return __salt__["ntp.delete_peers"](*peers, commit=False)


def _delete_ntp_servers(servers):
    """Calls ntp.delete_servers."""

    return __salt__["ntp.delete_servers"](*servers, commit=False)


def _exec_fun(name, *kargs):

    if name in list(globals().keys()):
        return globals().get(name)(*kargs)

    return None


def _check_diff_and_configure(fun_name, peers_servers, name="peers"):

    _ret = _default_ret(fun_name)

    _options = ["peers", "servers"]

    if name not in _options:
        return _ret

    _retrieve_fun = f"_retrieve_ntp_{name}"
    ntp_list_output = _exec_fun(
        _retrieve_fun
    )  # contains only IP Addresses as dictionary keys

    if ntp_list_output.get("result", False) is False:
        _ret["comment"] = "Cannot retrieve NTP {what} from the device: {reason}".format(
            what=name, reason=ntp_list_output.get("comment")
        )
        return _ret

    configured_ntp_list = set(ntp_list_output.get("out", {}))
    desired_ntp_list = set(peers_servers)

    if configured_ntp_list == desired_ntp_list:
        _ret.update(
            {
                "comment": f"NTP {name} already configured as needed.",
                "result": True,
            }
        )
        return _ret

    list_to_set = list(desired_ntp_list - configured_ntp_list)
    list_to_delete = list(configured_ntp_list - desired_ntp_list)

    list_to_set = _clean(list_to_set)
    list_to_delete = _clean(list_to_delete)

    changes = {}
    if list_to_set:
        changes["added"] = list_to_set
    if list_to_delete:
        changes["removed"] = list_to_delete

    _ret.update({"changes": changes})

    if __opts__["test"] is True:
        _ret.update(
            {"result": None, "comment": "Testing mode: configuration was not changed!"}
        )
        return _ret

    # <---- Retrieve existing NTP peers and determine peers to be added/removed --------------------------------------->

    # ----- Call _set_ntp_peers and _delete_ntp_peers as needed ------------------------------------------------------->

    expected_config_change = False
    successfully_changed = True

    comment = ""

    if list_to_set:
        _set_fun = f"_set_ntp_{name}"
        _set = _exec_fun(_set_fun, list_to_set)
        if _set.get("result"):
            expected_config_change = True
        else:  # something went wrong...
            successfully_changed = False
            comment += "Cannot set NTP {what}: {reason}".format(
                what=name, reason=_set.get("comment")
            )

    if list_to_delete:
        _delete_fun = f"_delete_ntp_{name}"
        _removed = _exec_fun(_delete_fun, list_to_delete)
        if _removed.get("result"):
            expected_config_change = True
        else:  # something went wrong...
            successfully_changed = False
            comment += "Cannot remove NTP {what}: {reason}".format(
                what=name, reason=_removed.get("comment")
            )

    _ret.update(
        {
            "successfully_changed": successfully_changed,
            "expected_config_change": expected_config_change,
            "comment": comment,
        }
    )

    return _ret


# ----------------------------------------------------------------------------------------------------------------------
# callable functions
# ----------------------------------------------------------------------------------------------------------------------


def managed(name, peers=None, servers=None):
    """
    Manages the configuration of NTP peers and servers on the device, as specified in the state SLS file.
    NTP entities not specified in these lists will be removed whilst entities not configured on the device will be set.

    SLS Example:

    .. code-block:: yaml

        netntp_example:
            netntp.managed:
                 - peers:
                    - 192.168.0.1
                    - 172.17.17.1
                 - servers:
                    - 24.124.0.251
                    - 138.236.128.36

    Output example:

    .. code-block:: python

        {
            'edge01.nrt04': {
                'netntp_|-netntp_example_|-netntp_example_|-managed': {
                    'comment': 'NTP servers already configured as needed.',
                    'name': 'netntp_example',
                    'start_time': '12:45:24.056659',
                    'duration': 2938.857,
                    'changes': {
                        'peers': {
                            'removed': [
                                '192.168.0.2',
                                '192.168.0.3'
                            ],
                            'added': [
                                '192.168.0.1',
                                '172.17.17.1'
                            ]
                        }
                    },
                    'result': None
                }
            }
        }
    """

    ret = _default_ret(name)
    result = ret.get("result", False)
    comment = ret.get("comment", "")
    changes = ret.get("changes", {})

    if not (
        isinstance(peers, list) or isinstance(servers, list)
    ):  # none of the is a list
        return ret  # just exit

    if isinstance(peers, list) and not _check(peers):  # check and clean peers
        ret["comment"] = (
            "NTP peers must be a list of valid IP Addresses or Domain Names"
        )
        return ret

    if isinstance(servers, list) and not _check(servers):  # check and clean servers
        ret["comment"] = (
            "NTP servers must be a list of valid IP Addresses or Domain Names"
        )
        return ret

    # ----- Retrieve existing NTP peers and determine peers to be added/removed --------------------------------------->

    successfully_changed = True
    expected_config_change = False

    if isinstance(peers, list):
        _peers_ret = _check_diff_and_configure(name, peers, name="peers")
        expected_config_change = _peers_ret.get("expected_config_change", False)
        successfully_changed = _peers_ret.get("successfully_changed", True)
        result = result and _peers_ret.get("result", False)
        comment += "\n" + _peers_ret.get("comment", "")
        _changed_peers = _peers_ret.get("changes", {})
        if _changed_peers:
            changes["peers"] = _changed_peers
    if isinstance(servers, list):
        _servers_ret = _check_diff_and_configure(name, servers, name="servers")
        expected_config_change = expected_config_change or _servers_ret.get(
            "expected_config_change", False
        )
        successfully_changed = successfully_changed and _servers_ret.get(
            "successfully_changed", True
        )
        result = result and _servers_ret.get("result", False)
        comment += "\n" + _servers_ret.get("comment", "")
        _changed_servers = _servers_ret.get("changes", {})
        if _changed_servers:
            changes["servers"] = _changed_servers

    ret.update({"changes": changes})

    if not (changes or expected_config_change):
        ret.update({"result": True, "comment": "Device configured properly."})
        return ret

    if __opts__["test"] is True:
        ret.update(
            {
                "result": None,
                "comment": (
                    "This is in testing mode, the device configuration was not changed!"
                ),
            }
        )
        return ret

    # <---- Call _set_ntp_peers and _delete_ntp_peers as needed --------------------------------------------------------

    # ----- Try to commit changes ------------------------------------------------------------------------------------->

    if expected_config_change:  # commit only in case there's something to update
        config_result, config_comment = __salt__["net.config_control"]()
        result = config_result and successfully_changed
        comment += config_comment

    # <---- Try to commit changes --------------------------------------------------------------------------------------

    ret.update({"result": result, "comment": comment})

    return ret



"""
Manage Windows Package Repository
"""

import itertools

# Python Libs
import os
import stat

import salt.config

# Salt Modules
import salt.runner
import salt.syspaths
import salt.utils.path


def __virtual__():
    return "winrepo"


def genrepo(name, force=False, allow_empty=False):
    """
    Refresh the winrepo.p file of the repository (salt-run winrepo.genrepo)

    If ``force`` is ``True`` no checks will be made and the repository will be
    generated if ``allow_empty`` is ``True`` then the state will not return an
    error if there are 0 packages,

    .. note::

        This state only loads on minions that have the ``roles: salt-master``
        grain set.

    Example:

    .. code-block:: yaml

        winrepo:
          winrepo.genrepo
    """

    ret = {"name": name, "result": True, "changes": {}, "comment": ""}

    master_config = salt.config.master_config(
        os.path.join(salt.syspaths.CONFIG_DIR, "master")
    )

    winrepo_dir = master_config["winrepo_dir"]
    winrepo_cachefile = master_config["winrepo_cachefile"]

    # We're actually looking for the full path to the cachefile here, so
    # prepend the winrepo_dir
    winrepo_cachefile = os.path.join(winrepo_dir, winrepo_cachefile)

    # Check if the winrepo directory exists
    # if not search for a file with a newer mtime than the winrepo_cachefile file
    execute = False
    if not force:
        if not os.path.exists(winrepo_dir):
            ret["result"] = False
            ret["comment"] = f"{winrepo_dir} is missing"
            return ret
        elif not os.path.exists(winrepo_cachefile):
            execute = True
            ret["comment"] = f"{winrepo_cachefile} is missing"
        else:
            winrepo_cachefile_mtime = os.stat(winrepo_cachefile)[stat.ST_MTIME]
            for root, dirs, files in salt.utils.path.os_walk(winrepo_dir):
                for name in itertools.chain(files, dirs):
                    full_path = os.path.join(root, name)
                    if os.stat(full_path)[stat.ST_MTIME] > winrepo_cachefile_mtime:
                        ret["comment"] = "mtime({}) < mtime({})".format(
                            winrepo_cachefile, full_path
                        )
                        execute = True
                        break

    if __opts__["test"]:
        ret["result"] = None
        return ret

    if not execute and not force:
        return ret

    runner = salt.runner.RunnerClient(master_config)
    runner_ret = runner.cmd("winrepo.genrepo", [])
    ret["changes"] = {"winrepo": runner_ret}
    if isinstance(runner_ret, dict) and runner_ret == {} and not allow_empty:
        os.remove(winrepo_cachefile)
        ret["result"] = False
        ret["comment"] = "winrepo.genrepo returned empty"
    return ret
